from sys import exc_info
from unittest import TestCase
from warnings import warn


class GcAfterTestTests(TestCase):
    def tearDown(self):
        try:
            del self.cycle
        except AttributeError:
            pass

    def test_okay(self):
        pass

    def test_cycle_without_resource(self):
        self.cycle = _Cycle()

    def test_cycle_with_resource(self):
        self.cycle = _Cycle(resource=_Resource())

    def test_test_holds_cycle(self):
        self.hold_cycle = _Cycle(resource=_Resource())

    def test_failure(self):
        raise AssertionError("failure")

    def test_exception(self):
        1 / 0

    def test_traceback_cycle(self):

        def f():
            try:
                1 / 0
            except Exception:
                # create cycle
                tb = exc_info()[2]  # noqa: F841

        f()


class _Cycle:
    """Auxiliary class creating a reference cycle."""

    def __init__(self, **kw):
        self.self = self  # create reference cycle
        self.__dict__.update(kw)


class _Resource:
    """Auxiliary class emulating a resource."""
    closed = False

    def close(self):
        self.closed = True

    def __del__(self):
        if not self.closed:
            warn(ResourceWarning(
                "not closed"
                " - this is no error: testing ResourceWarning here"))



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest
import gc


# for some early 3.13 versions, the third threshold has been
# a constant 0; this was changed again for newer versions
_thresholds = gc.get_threshold()
gc.set_threshold(10, 10, 10)
ZERO_THR3 = gc.get_threshold()[2] == 0
gc.set_threshold(*_thresholds)


def make_sure_gc_threshold_is_701_11_9():
    pass


make_sure_gc_threshold_is_701_11_9.__doc__ = """\
>>> import gc
>>> gc.get_threshold()
(701, 11, %d)
""" % (0 if ZERO_THR3 else 9)


def test_suite():
    return doctest.DocTestSuite()



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Sample tests with a layer that produces output on stderr
"""

import doctest
import sys
import unittest


class Layer:

    def setUp(self):
        pass
    setUp = classmethod(setUp)

    def tearDown(self):
        pass
    tearDown = classmethod(tearDown)


def test_something():
    """
    >>> 1 + 1
    2
    """


def test_suite():
    # Generate some text on stderr to be sure the test runner can handle it.
    sys.stderr.write('A message on stderr.'
                     '  Please ignore (expected in test output).\n')

    suite = unittest.TestSuite()
    d = doctest.DocTestSuite()
    d.layer = Layer
    suite.addTest(d)
    return suite


if __name__ == '__main__':
    unittest.main()



##############################################################################
#
# Copyright (c) 2004-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Command-line option parsing
"""

import argparse
import os
import re
import sys
from importlib.metadata import distribution

from zope.testrunner.formatter import ColorfulOutputFormatter
from zope.testrunner.formatter import OutputFormatter
from zope.testrunner.formatter import SubunitOutputFormatter
from zope.testrunner.formatter import SubunitV2OutputFormatter
from zope.testrunner.formatter import terminal_has_colors
from zope.testrunner.profiling import available_profilers

from .util import uses_refcounts


def _regex_search(s):
    return re.compile(s).search


parser = argparse.ArgumentParser(
    description="Discover and run unittest tests")


parser.add_argument("legacy_module_filter", nargs="?",
                    help="DEPRECATED: Prefer to use --module.")


parser.add_argument("legacy_test_filter", nargs="?",
                    help="DEPRECATED: Prefer to use --test.")

######################################################################
# Searching and filtering

searching = parser.add_argument_group("Searching and filtering", """\
Options in this group are used to define which tests to run.
""")

searching.add_argument(
    '--package', '--dir', '-s', action="append", dest='package',
    help="""\
Search the given package's directories for tests.  This can be
specified more than once to run tests in multiple parts of the source
tree.  For example, if refactoring interfaces, you don't want to see
the way you have broken setups for tests in other packages. You *just*
want to run the interface tests.

Packages are supplied as dotted names.  For compatibility with the old
test runner, forward and backward slashed in package names are
converted to dots.

(In the special case of packages spread over multiple directories,
only directories within the test search path are searched. See the
--path option.)

""")

searching.add_argument(
    '--module', '-m', action="append", dest='module',
    help="""\
Specify a test-module filter as a regular expression.  This is a
case-sensitive regular expression, used in search (not match) mode, to
limit which test modules are searched for tests.  The regular
expressions are checked against dotted module names.  In an extension
of Python regexp notation, a leading "!" is stripped and causes the
sense of the remaining regexp to be negated (so "!bc" matches any
string that does not match "bc", and vice versa).  The option can be
specified multiple test-module filters.  Test modules matching any of
the test filters are searched.  If no test-module filter is specified,
then all test modules are used.
""")

searching.add_argument(
    '--test', '-t', action="append", dest='test',
    help="""\
Specify a test filter as a regular expression.  This is a
case-sensitive regular expression, used in search (not match) mode, to
limit which tests are run.  In an extension of Python regexp notation,
a leading "!" is stripped and causes the sense of the remaining regexp
to be negated (so "!bc" matches any string that does not match "bc",
and vice versa).  The option can be specified multiple test filters.
Tests matching any of the test filters are included.  If no test
filter is specified, then all tests are run.
""")

searching.add_argument(
    '--unit', '-u', action="store_true", dest='unit',
    help="""\
Run only unit tests, ignoring any layer options.
""")

searching.add_argument(
    '--non-unit', '-f', action="store_true", dest='non_unit',
    help="""\
Run tests other than unit tests.
""")

searching.add_argument(
    '--layer', action="append", dest='layer',
    help="""\
Specify a test layer to run.  The option can be given multiple times
to specify more than one layer.  If not specified, all layers are run.
It is common for the running script to provide default values for this
option.  Layers are specified regular expressions, used in search
mode, for dotted names of objects that define a layer.  In an
extension of Python regexp notation, a leading "!" is stripped and
causes the sense of the remaining regexp to be negated (so "!bc"
matches any string that does not match "bc", and vice versa).  The
layer named 'zope.testrunner.layer.UnitTests' is reserved for
unit tests, however, take note of the --unit and non-unit options.
""")

searching.add_argument(
    '-a', '--at-level', type=int, dest='at_level',
    default=1,
    help="""\
Run the tests at the given level.  Any test at a level at or below
this is run, any test at a level above this is not run.  Level <= 0
runs all tests.
""")

searching.add_argument(
    '--all', action="store_true", dest='all',
    help="Run tests at all levels.")

searching.add_argument(
    '--only-level', type=int, dest='only_level',
    default=None,
    help="""\
Run only the tests at the given level. Tests on other levels are not run.
Default is None, which means the level is not restricted by this option.
If this option is used it overrides `--at-level` and `--all` options.
""")

searching.add_argument(
    '--list-tests', action="store_true", dest='list_tests',
    default=False,
    help="List all tests that matched your filters.  Do not run any tests.")

searching.add_argument(
    '--require-unique', action="store_true", dest='require_unique_ids',
    default=False,
    help="""\
Require that all test IDs be unique and raise an error if duplicates are
encountered.
""")


######################################################################
# Reporting

reporting = parser.add_argument_group("Reporting", """\
Reporting options control basic aspects of test-runner output
""")

reporting.add_argument(
    '--verbose', '-v', action="count", dest='verbose',
    default=0,
    help="""\
Make output more verbose.
Increment the verbosity level.
""")

reporting.add_argument(
    '--quiet', '-q', action="store_true", dest='quiet',
    help="""\
Make the output minimal, overriding any verbosity options.
""")

reporting.add_argument(
    '--progress', '-p', action="store_true", dest='progress',
    help="""\
Output progress status
""")

reporting.add_argument(
    '--no-progress', action="store_false", dest='progress',
    help="""\
Do not output progress status.  This is the default, but can be used to
counter a previous use of --progress or -p.
""")

# The actual processing will be done in the get_options function, but
# we want argparse to generate appropriate help info for us, so we add
# an option anyway.
reporting.add_argument(
    '--auto-progress', action="store_const", const=None,
    help="""\
Output progress status, but only when stdout is a terminal.
""")

reporting.add_argument(
    '--color', '-c', action="store_true", dest='color',
    help="""\
Colorize the output.
""")

reporting.add_argument(
    '--no-color', '-C', action="store_false", dest='color',
    help="""\
Do not colorize the output.  This is the default, but can be used to
counter a previous use of --color or -c.
""")

# The actual processing will be done in the get_options function, but
# we want argparse to generate appropriate help info for us, so we add
# an option anyway.
reporting.add_argument(
    '--auto-color', action="store_const", const=None,
    help="""\
Colorize the output, but only when stdout is a terminal.
""")

reporting.add_argument(
    '--subunit', action="store_true", dest='subunit',
    help="""\
Use subunit v1 output.  Will not be colorized.
""")

reporting.add_argument(
    '--subunit-v2', action="store_true", dest='subunit_v2',
    help="""\
Use subunit v2 output.  Will not be colorized.
""")

reporting.add_argument(
    '--slow-test', type=float, dest='slow_test_threshold', metavar='N',
    default=10,
    help="""\
With -c and -vvv, highlight tests that take longer than N seconds (default:
%(default)s).
""")

reporting.add_argument(
    '-1', '--hide-secondary-failures',
    action="store_true", dest='report_only_first_failure',
    help="""\
Report only the first failure in a doctest. (Examples after the
failure are still executed, in case they do any cleanup.)
""")

reporting.add_argument(
    '--show-secondary-failures',
    action="store_false", dest='report_only_first_failure',
    help="""\
Report all failures in a doctest.  This is the default, but can
be used to counter a default use of -1 or --hide-secondary-failures.
""")

reporting.add_argument(
    '--ndiff', action="store_true", dest="ndiff",
    help="""\
When there is a doctest failure, show it as a diff using the ndiff.py utility.
""")

reporting.add_argument(
    '--udiff', action="store_true", dest="udiff",
    help="""\
When there is a doctest failure, show it as a unified diff.
""")

reporting.add_argument(
    '--cdiff', action="store_true", dest="cdiff",
    help="""\
When there is a doctest failure, show it as a context diff.
""")

reporting.add_argument(
    '--ignore-new-thread',
    metavar='REGEXP',
    action="append",
    default=[],
    dest='ignore_new_threads',
    help="""\
If a thread with this name is left behind, don't report this at the end.
This is a case-sensitive regular expression, used in match mode.
This option can be used multiple times. If a thread name matches any of them,
it will be ignored.
""")

reporting.add_argument(
    '--buffer', action="store_true", dest="buffer",
    help="""\
Buffer the standard output and standard error streams during each test.
Output during a passing test is discarded.
Output during failing or erroring tests is echoed.
This option is enabled by default if --subunit or --subunit-v2 is in use, to
avoid corrupting the subunit stream.
""")

reporting.add_argument(
    '--xml', action='store', dest='xmlOutput', help="""\
If given, XML reports will be written to the specified directory.
""")


######################################################################
# Analysis

analysis = parser.add_argument_group("Analysis", """\
Analysis options provide tools for analysing test output.
""")


analysis.add_argument(
    '--stop-on-error', '--stop', '-x', action="store_true",
    dest='stop_on_error',
    help="Stop running tests after first test failure or error."
)

analysis.add_argument(
    '--post-mortem', '--pdb', '-D', action="store_true", dest='post_mortem',
    help="Enable post-mortem debugging of test failures"
)


analysis.add_argument(
    '--gc', '-g', action="append", dest='gc', type=int,
    help="""\
Set the garbage collector generation threshold.  This can be used
to stress memory and gc correctness.  Some crashes are only
reproducible when the threshold is set to 1 (aggressive garbage
collection).  Do "--gc 0" to disable garbage collection altogether.

The --gc option can be used up to 3 times to specify up to 3 of the 3
Python gc_threshold settings.

""")

analysis.add_argument(
    '--gc-option', '-G', action="append", dest='gc_option',
    choices={'DEBUG_STATS', 'DEBUG_COLLECTABLE', 'DEBUG_UNCOLLECTABLE',
             'DEBUG_INSTANCES', 'DEBUG_OBJECTS', 'DEBUG_SAVEALL',
             'DEBUG_LEAK'},
    help="""\
Set a Python gc-module debug flag.  This option can be used more than
once to set multiple flags.
""")

if uses_refcounts:
    analysis.add_argument(
        '--gc-after-test', action="store_true", dest='gc_after_test',
        help="""\
    After each test, call 'gc.collect' and record the return
    value *rv*; when *rv* is non-zero, output '!' on verbosity level 1
    and '[*rv*]' on higher verbosity levels.\n
    On verbosity level 4 or higher output detailed cycle information.
    """)

analysis.add_argument(
    '--repeat', '-N', action="store", type=int, dest='repeat',
    default=1,
    help="""\
Repeat the tests the given number of times.  This option is used to
make sure that tests leave their environment in the state they found
it and, with the --report-refcounts option to look for memory leaks.
""")

analysis.add_argument(
    '--report-refcounts', '-r', action="store_true", dest='report_refcounts',
    help="""\
After each run of the tests, output a report summarizing changes in
refcounts by object type.  This option that requires that Python was
built with the --with-pydebug option to configure.
""")

analysis.add_argument(
    '--coverage', action="store", dest='coverage',
    help="""\
Perform code-coverage analysis, saving trace data to the directory
with the given name.  A code coverage summary is printed to standard
out.
""")

analysis.add_argument(
    '--profile', action="store", dest='profile',
    choices=set(available_profilers),
    help="""\
Run the tests under cProfiler and display the top 50 stats, sorted
by cumulative time and number of calls.
""")
analysis.add_argument(
    '--profile-directory', action="store", dest='prof_dir', default='.',
    help="""\
Directory for temporary profiler files.  All files named tests_profile.*.prof
in this directory will be removed.  If you intend to run multiple instances
of the test runner in parallel, be sure to tell them to use different
directories, so they won't step on each other's toes.
""")

######################################################################
# Setup

setup = parser.add_argument_group("Setup", """\
Setup options are normally supplied by the testrunner script, although
they can be overridden by users.
""")

setup.add_argument(
    '--path', action="append", dest='path',
    type=os.path.abspath,
    help="""\
Specify a path to be added to Python's search path.  This option can
be used multiple times to specify multiple search paths.  The path is
usually specified by the test-runner script itself, rather than by
users of the script, although it can be overridden by users.  Only
tests found in the path will be run.

This option also specifies directories to be searched for tests.
See the search_directory.
""")

setup.add_argument(
    '--test-path', action="append", dest='test_path',
    type=os.path.abspath,
    help="""\
Specify a path to be searched for tests, but not added to the Python
search path.  This option can be used multiple times to specify
multiple search paths.  The path is usually specified by the
test-runner script itself, rather than by users of the script,
although it can be overridden by users.  Only tests found in the path
will be run.
""")

setup.add_argument(
    '--package-path', action="append", dest='package_path', nargs=2,
    metavar="ARG",
    help="""\
Specify a path to be searched for tests, but not added to the Python
search path.  Also specify a package for files found in this path.
This is used to deal with directories that are stitched into packages
that are not otherwise searched for tests.

This option takes 2 arguments specifying the path and the package.

This option can be used multiple times to specify
multiple search paths.  The path is usually specified by the
test-runner script itself, rather than by users of the script,
although it can be overridden by users.  Only tests found in the path
will be run.
""")

setup.add_argument(
    '--tests-pattern', action="store", dest='tests_pattern',
    default=_regex_search('^tests$'),
    type=_regex_search,
    help="""\
The test runner looks for modules containing tests.  It uses this
pattern to identify these modules.  The modules may be either packages
or python files.

If a test module is a package, it uses the value given by the
test-file-pattern to identify python files within the package
containing tests.
""")

setup.add_argument(
    '--suite-name', action="store", dest='suite_name',
    default='test_suite',
    help="""\
Specify the name of the object in each test_module that contains the
module's test suite.
""")

setup.add_argument(
    '--test-file-pattern', action="store", dest='test_file_pattern',
    default=_regex_search('^test'),
    type=_regex_search,
    help="""\
Specify a pattern for identifying python files within a tests package.
See the documentation for the --tests-pattern option.
""")

setup.add_argument(
    '--ignore_dir', action="append", dest='ignore_dir',
    default=['.git', '.svn', 'CVS', '{arch}', '.arch-ids', '_darcs'],
    help="""\
Specifies the name of a directory to ignore when looking for tests.
""")

setup.add_argument(
    '--shuffle', action="store_true", dest='shuffle',
    help="""\
Shuffles the order in which tests are ran.
""")

setup.add_argument(
    '--shuffle-seed', action="store", dest='shuffle_seed', type=int,
    help="""\
Value used to initialize the tests shuffler. Specify a value to create
repeatable random ordered tests.
""")


######################################################################
# Other

other = parser.add_argument_group("Other", "Other options")

other.add_argument(
    '--version', action="store_true", dest='showversion',
    help="Print the version of the testrunner, and exit.")

other.add_argument(
    '-j', action="store", type=int, dest='processes',
    default=1,
    help="""\
Use up to given number of parallel processes to execute tests.  May decrease
test run time substantially.  Defaults to %(default)s.
""")

other.add_argument(
    '--keepbytecode', '-k', action="store_true", dest='keepbytecode',
    help="""\
Normally, the test runner scans the test paths and the test
directories looking for and deleting pyc or pyo files without
corresponding py files.  This is to prevent spurious test failures due
to finding compiled modules where source modules have been deleted.
This scan can be time consuming.  Using this option disables this
scan.  If you know you haven't removed any modules since last running
the tests, can make the test run go much faster.
""")

other.add_argument(
    '--usecompiled', action="store_true", dest='usecompiled',
    help="""\
Normally, a package must contain an __init__.py file, and only .py files
can contain test code.  When this option is specified, compiled Python
files (.pyc and .pyo) can be used instead:  a directory containing
__init__.pyc or __init__.pyo is also considered to be a package, and if
file XYZ.py contains tests but is absent while XYZ.pyc or XYZ.pyo exists
then the compiled files will be used.  This is necessary when running
tests against a tree where the .py files have been removed after
compilation to .pyc/.pyo.  Use of this option implies --keepbytecode.
""")

other.add_argument(
    '--exit-with-status', action="store_true", dest='exitwithstatus',
    help="""DEPRECATED: The test runner will always exit with a status.\
""")


######################################################################
# Command-line processing


def merge_options(options, defaults):
    odict = options.__dict__
    for name, value in defaults.__dict__.items():
        if (value is not None) and (odict[name] is None):
            odict[name] = value


def get_options(args=None, defaults=None):
    # Because we want to inspect stdout and decide to colorize or not, we
    # replace the --auto-color option with the appropriate --color or
    # --no-color option.  That way the subprocess doesn't have to decide (which
    # it would do incorrectly anyway because stdout would be a pipe).
    def apply_auto_color(args):
        if args and '--auto-color' in args:
            if sys.stdout.isatty() and terminal_has_colors():
                colorization = '--color'
            else:
                colorization = '--no-color'

            args[:] = [arg.replace('--auto-color', colorization)
                       for arg in args]

    # The comment of apply_auto_color applies here as well
    def apply_auto_progress(args):
        if args and '--auto-progress' in args:
            if sys.stdout.isatty():
                progress = '--progress'
            else:
                progress = '--no-progress'

            args[:] = [arg.replace('--auto-progress', progress)
                       for arg in args]

    apply_auto_color(args)
    apply_auto_color(defaults)
    apply_auto_progress(args)
    apply_auto_progress(defaults)

    if defaults:
        defaults = parser.parse_args(defaults)
    else:
        defaults = None

    if args is None:
        args = sys.argv

    options = parser.parse_args(args[1:], defaults)
    options.original_testrunner_args = args

    if options.showversion:
        dist = distribution('zope.testrunner')
        print('zope.testrunner version %s' % dist.version)
        sys.exit(0)

    if options.subunit or options.subunit_v2:
        try:
            import subunit
            subunit
        except ImportError:
            print("""\
        Subunit is not installed. Please install Subunit
        to generate subunit output.
        """)
            options.fail = True
            return options
        options.buffer = True

    if options.subunit and options.subunit_v2:
        print("""\
        You may only use one of --subunit and --subunit-v2.
        """)
        options.fail = True
        return options

    if options.subunit:
        options.output = SubunitOutputFormatter(options)
    elif options.subunit_v2:
        options.output = SubunitV2OutputFormatter(options)
    elif options.color:
        options.output = ColorfulOutputFormatter(options)
        options.output.slow_test_threshold = options.slow_test_threshold
    else:
        options.output = OutputFormatter(options)

    options.fail = False

    if options.legacy_module_filter:
        module_filter = options.legacy_module_filter
        if module_filter != '.':
            if options.module:
                options.module.append(module_filter)
            else:
                options.module = [module_filter]

        if options.legacy_test_filter:
            test_filter = options.legacy_test_filter
            if options.test:
                options.test.append(test_filter)
            else:
                options.test = [test_filter]

    options.ignore_dir = set(options.ignore_dir)
    options.test = options.test or ['.']
    module_set = bool(options.module)
    options.module = options.module or ['.']

    options.path = options.path or []
    options.test_path = options.test_path or []
    options.test_path += options.path

    options.test_path = ([(path, '') for path in options.test_path]
                         +
                         [(os.path.abspath(path), package)
                          for (path, package) in options.package_path or ()
                          ])

    if options.package:
        pkgmap = dict(options.test_path)
        options.package = [normalize_package(p, pkgmap)
                           for p in options.package]

    options.prefix = [(path + os.path.sep, package)
                      for (path, package) in options.test_path]

    # Sort prefixes so that longest prefixes come first.
    # That is because only first match is evaluated which
    # can be a problem with nested source packages.
    options.prefix.sort(key=lambda p: len(p[0]), reverse=True)

    if options.all:
        options.at_level = sys.maxsize

    if options.unit and options.non_unit:
        # The test runner interprets this as "run only those tests that are
        # both unit and non-unit at the same time".  The user, however, wants
        # to run both unit and non-unit tests.  Disable the filtering so that
        # the user will get what she wants:
        options.unit = options.non_unit = False

    if options.unit:
        # XXX Argh.
        options.layer = ['zope.testrunner.layer.UnitTests']

    options.layer = options.layer and {layer: 1 for layer in options.layer}

    if options.usecompiled:
        options.keepbytecode = options.usecompiled

    if options.quiet:
        options.verbose = 0

    if options.report_refcounts and options.repeat < 2:
        print("""\
        You must use the --repeat (-N) option to specify a repeat
        count greater than 1 when using the --report_refcounts (-r)
        option.
        """)
        options.fail = True
        return options

    if options.report_refcounts and not hasattr(sys, "gettotalrefcount"):
        print("""\
        The Python you are running was not configured
        with --with-pydebug. This is required to use
        the --report-refcounts option.
        """)
        options.fail = True
        return options

    if module_set and options.require_unique_ids:
        # We warn if --module and --require-unique are specified at the same
        # time, though we don't exit.
        print("""\
        You specified a module along with --require-unique;
        --require-unique will not try to enforce test ID uniqueness when
        working with a specific module.
        """)

    return options


def normalize_package(package, package_map=None):
    r"""Normalize package name passed to the --package option.

        >>> normalize_package('zope.testrunner')
        'zope.testrunner'

    Converts path names into package names for compatibility with the old
    test runner.

        >>> normalize_package('zope/testrunner')
        'zope.testrunner'
        >>> normalize_package('zope/testrunner/')
        'zope.testrunner'
        >>> normalize_package('zope\\testrunner')
        'zope.testrunner'

    Can use a map of absolute pathnames to package names

        >>> a = os.path.abspath
        >>> normalize_package('src/zope/testrunner/',
        ...                   {a('src'): ''})
        'zope.testrunner'
        >>> normalize_package('src/zope_testrunner/',
        ...                   {a('src/zope_testrunner'): 'zope.testrunner'})
        'zope.testrunner'
        >>> normalize_package('src/zope_something/tests',
        ...                   {a('src/zope_something'): 'zope.something',
        ...                    a('src'): ''})
        'zope.something.tests'

    """
    package_map = {} if package_map is None else package_map
    package = package.replace('\\', '/')
    if package.endswith('/'):
        package = package[:-1]
    bits = package.split('/')
    for n in range(len(bits), 0, -1):
        pkg = package_map.get(os.path.abspath('/'.join(bits[:n])))
        if pkg is not None:
            bits = bits[n:]
            if pkg:
                bits = [pkg] + bits
            return '.'.join(bits)
    return package.replace('/', '.')



##############################################################################
#
# Copyright (c) 2012-2018 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Sample tests with layers that have broken set up and tear down."""

import unittest


class BrokenSetUpLayer:

    @classmethod
    def setUp(cls):
        raise ValueError('No value is good enough for me!')

    @classmethod
    def tearDown(cls):
        pass


class BrokenTearDownLayer:

    @classmethod
    def setUp(cls):
        pass

    @classmethod
    def tearDown(cls):
        raise TypeError('You are not my type.  No-one is my type!')


class TestSomething1(unittest.TestCase):

    layer = BrokenSetUpLayer

    def test_something(self):
        pass


class TestSomething2(unittest.TestCase):

    layer = BrokenTearDownLayer

    def test_something(self):
        pass


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestSomething1))
    suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestSomething2))
    return suite



##############################################################################
#
# Copyright (c) 2004-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test runner interfaces

XXX Note: These interfaces are still being sketched out. Please do not rely on
them, yet.

"""

import zope.interface


class EndRun(Exception):
    """Indicate that the existing run call should stop

    Used to prevent additional test output after post-mortem debugging.

    """


class IFeature(zope.interface.Interface):
    """Features extend the test runners functionality in a pipe-lined
    order.
    """

    active = zope.interface.Attribute(
        "Flag whether this feature is activated. If it is not activated than "
        "its methods won't be called by the runner.")

    def global_setup():
        """Executed once when the test runner is being set up."""

    def late_setup():
        """Executed once right before the actual tests get executed and after
        all global setups have happened.

        Should do as little work as possible to avoid timing interferences
        with other features.

        It is guaranteed that the calling stack frame is not left until
        early_teardown was called.

        """

    def layer_setup(layer):
        """Executed once after a layer was set up."""

    def layer_teardown(layer):
        """Executed once after a layer was run."""

    def test_setup(test):
        """Executed once before each test."""

    def test_teardown(test):
        """Executed once after each test."""

    def early_teardown():
        """Executed once directly after all tests.

        This method should do as little as possible to avoid timing issues.

        It is guaranteed to be called directly from the same stack frame that
        called `late_setup`.

        """

    def global_teardown():
        """Executed once after all tests where run and early teardowns have
        happened.

        """

    def report():
        """Executed once after all tests have been run and all setup was
        torn down.

        This is the only method that should produce output.

        """


class ITestRunner(zope.interface.Interface):
    """The test runner manages test layers and their execution.

    The functionality of a test runner can be extended by configuring
    features.

    """

    options = zope.interface.Attribute(
        "Provides access to configuration options.")


class IMinimalTestLayer(zope.interface.Interface):
    """A test layer.

    This is the bare minimum that a ``layer`` attribute on a test suite
    should provide.
    """

    __bases__ = zope.interface.Attribute(
        "A tuple of base layers.")

    __name__ = zope.interface.Attribute(
        "Name of the layer")

    __module__ = zope.interface.Attribute(
        "Dotted name of the module that defines this layer")

    def __hash__():
        """A test layer must be hashable.

        The default identity-based __eq__ and __hash__ that Python provides
        usually suffice.
        """

    def __eq__():
        """A test layer must be hashable.

        The default identity-based __eq__ and __hash__ that Python provides
        usually suffice.
        """


class IFullTestLayer(IMinimalTestLayer):
    """A test layer.

    This is the full list of optional methods that a test layer can specify.
    """

    def setUp():
        """Shared layer setup.

        Called once before any of the tests in this layer (or sublayers)
        are run.
        """

    def tearDown():
        """Shared layer teardown.

        Called once after all the tests in this layer (and sublayers)
        are run.

        May raise NotImplementedError.
        """

    def testSetUp():
        """Additional test setup.

        Called once before every of test in this layer (or sublayers)
        is run.
        """

    def testTearDown():
        """Additional test teardown.

        Called once after every of test in this layer (or sublayers)
        is run.
        """



##############################################################################
#
# Copyright (c) 2013 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import unittest


class TestSkipppedWithLayer(unittest.TestCase):
    layer = "sample_skipped_tests.Layer"

    @unittest.skip('Hop, skipped')
    def test_layer_skipped(self):
        pass

    def test_layer_pass(self):
        pass


class Layer:
    pass


class TestSkipppedNoLayer(unittest.TestCase):
    # Only one test, so we don't have to bother about ordering when the runner
    # displays the tests' list while in verbose mode.

    @unittest.skip("I'm a skipped test!")
    def test_skipped(self):
        pass



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import time
import unittest


class ClassicLeakable:
    def __init__(self):
        self.x = 'x'


class Leakable:
    def __init__(self):
        self.x = 'x'


leaked = []


class TestSomething(unittest.TestCase):

    def testleak(self):
        leaked.append((ClassicLeakable(), Leakable(), time.time()))


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(
        unittest.defaultTestLoader.loadTestsFromTestCase(TestSomething))
    return suite


if __name__ == '__main__':
    unittest.main()



##############################################################################
#
# Copyright (c) 2022 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Some general auxiliary functions.
"""
import platform
import sys


is_jython = sys.platform.startswith('java')
is_pypy = platform.python_implementation() == "PyPy"

uses_refcounts = not (is_jython or is_pypy)



##############################################################################
#
# Copyright (c) 2004-2013 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test runner
"""

import os
import sys

import pkg_resources  # noqa: F401 unused import, see GitHub issue #194


def run(defaults=None, args=None, script_parts=None, cwd=None, warnings=None):
    """Main runner function which can be and is being used from main programs.

    Will execute the tests and exit the process according to the test result.

    .. versionchanged:: 4.8.0
       Add the *warnings* keyword argument.
       See :class:`zope.testrunner.runner.Runner`

    """
    failed = run_internal(defaults, args, script_parts=script_parts, cwd=cwd,
                          warnings=warnings)
    sys.exit(int(failed))


def run_internal(defaults=None, args=None, script_parts=None, cwd=None,
                 warnings=None):
    """Execute tests.

    Returns whether errors or failures occured during testing.

    .. versionchanged:: 4.8.0
       Add the *warnings* keyword argument.
       See :class:`zope.testrunner.runner.Runner`

    """
    if script_parts is None:
        script_parts = _script_parts(args)
    if cwd is None:
        cwd = os.getcwd()
    # XXX Bah. Lazy import to avoid circular/early import problems
    from zope.testrunner.runner import Runner
    runner = Runner(
        defaults, args, script_parts=script_parts, cwd=cwd, warnings=warnings)
    runner.run()
    return runner.failed


def _script_parts(args=None):
    script_parts = (args or sys.argv)[0:1]
    # If we are running via setup.py, then we'll have to run the
    # sub-process differently.
    if script_parts[0] == 'setup.py':
        script_parts = ['-c', 'from zope.testrunner import run; run()']
    else:
        # make sure we remember the absolute path early -- the tests might
        # do an os.chdir()
        script_parts[0] = os.path.abspath(script_parts[0])
        if not os.path.exists(script_parts[0]):
            # uhh, no wrapper script?  this happens on Windows sometimes,
            # where there are like three wrapper scripts with various suffixes,
            # and I don't want to go looking up what they might be.
            script_parts = ['-m', 'zope.testrunner']
    return script_parts


if __name__ == '__main__':
    # this used to allow people to try out the test runner with
    #   python -m zope.testrunner --test-path .
    # on Python 2.5.  This broke on 2.6, and 2.7 and newer use __main__.py
    # for that.  But there are some users out there who actually use
    #   python -e zope.testrunner.__init__ --test-path .
    # so let's keep this for BBB
    run()



##############################################################################
#
# Copyright (c) 2004-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Test discovery
"""

import os
import re
import sys
import unittest

import zope.testrunner.debug
import zope.testrunner.feature
import zope.testrunner.layer
from zope.testrunner.filter import build_filtering_func


identifier = re.compile(r'[_a-z]\w*$', re.I).match
IGNORE_FOLDERS = {
    '.git',
    'node_modules',
    '__pycache__'
}


class DuplicateTestIDError(Exception):
    """Raised whenever a test ID is encountered twice during loading."""


class StartUpFailure(unittest.TestCase):
    """Empty test case added to the test suite to indicate import failures.

    >>> class Options(object):
    ...    post_mortem = False
    >>> options = Options()

    Normally the StartUpFailure just acts as an empty test suite to satisfy
    the test runner and statistics:

    >>> s = StartUpFailure(options, 'fauxmodule', None)
    >>> s
    <StartUpFailure module=fauxmodule>
    >>> isinstance(s,unittest.TestCase)
    True
    >>> s.shortDescription()
    'StartUpFailure: import errors in fauxmodule.'

    Sometimes test suited collected by zope.testrunner end up being run
    by a regular unittest TestRunner, and it's not hard to make sure
    StartUpFailure does something sensible in that case

    >>> r = unittest.TestResult()
    >>> _ = s.run(r)
    >>> print(r.failures[0][1].rstrip()) # doctest: +ELLIPSIS
    Traceback (most recent call last):
      ...
    AssertionError: could not import fauxmodule

    If you'd like more details, be sure to pass the original exc_info::

    >>> import sys
    >>> try:
    ...    raise Exception('something bad happened during import')
    ... except:
    ...     exc_info = sys.exc_info()

    >>> s = StartUpFailure(options, 'fauxmodule', exc_info)

    >>> r = unittest.TestResult()
    >>> _ = s.run(r)
    >>> print(r.errors[0][0].shortDescription())
    StartUpFailure: import errors in fauxmodule.
    >>> print(r.errors[0][1].rstrip()) # doctest: +ELLIPSIS
    Traceback (most recent call last):
      ...
    Exception: something bad happened during import

    However, if the post mortem option is enabled:

    >>> options.post_mortem = True

    ...then the the StartUpFailure will start the debugger and stop
    the test run after the debugger quits.

    To simulate the user pressing 'c' and hitting return in the
    debugger, we use a FakeInputContinueGenerator:

    >>> from zope.testrunner.runner import FakeInputContinueGenerator
    >>> old_stdin = sys.stdin
    >>> sys.stdin = FakeInputContinueGenerator()

    Now we can see the EndRun exception that is raised by the
    postmortem debugger to indicate that debugging is finished and the
    test run should be terminated:

    >>> from zope.testrunner.interfaces import EndRun
    >>> try: #doctest: +ELLIPSIS
    ...     # Needed to prevent the result from starting with '...'
    ...     print("Result:")
    ...     StartUpFailure(options, None, exc_info)
    ... except EndRun:
    ...     print("EndRun raised")
    ... finally:
    ...     sys.stdin = old_stdin
    Result:
    Exception: something bad happened during import
    ...
    (Pdb) c
    <BLANKLINE>
    **********************************************************************
    Can't use pdb.set_trace when running a layer as a subprocess!
    **********************************************************************
    <BLANKLINE>
    EndRun raised

    Annoyingly, sometimes StartUpFailures occur when postmortem debugging
    is enabled but no exc_info is passed. In this case, we raise a
    sensible exception rather than letting the debugger barf with an
    AttributeError:

    >>> options.post_mortem = True
    >>> StartUpFailure(options, None, exc_info[:2]+(None,))
    Traceback (most recent call last):
    ...
    TypeError: If post_mortem is specified, full exc_info must be passed!
    """

    def __init__(self, options, module, exc_info):
        if options.post_mortem:
            for item in exc_info:
                if item is None:
                    raise TypeError('If post_mortem is specified, '
                                    'full exc_info must be passed!')
            zope.testrunner.debug.post_mortem(exc_info)
        self.module = module
        self.exc_info = exc_info
        super().__init__()

    def shortDescription(self):
        return 'StartUpFailure: import errors in %s.' % self.module

    def __repr__(self):
        return '<StartUpFailure module=%s>' % self.module

    def runTest(self):
        if self.exc_info is None or any(x is None for x in self.exc_info):
            self.fail("could not import %s" % self.module)
        else:
            try:
                _, value, tb = self.exc_info
                raise value.with_traceback(tb)
            finally:
                value = None
                tb = None


def find_tests(options, found_suites=None):
    """Creates a dictionary mapping layer name to a suite of tests to be run
    in that layer.

    Passing a list of suites using the found_suites parameter will cause
    that list of suites to be used instead of attempting to load them from
    the filesystem. This is useful for unit testing the test runner.
    """
    remove_stale_bytecode(options)
    suites = {}
    dupe_ids = set()

    test_accept = build_filtering_func(options.test)
    module_accept = build_filtering_func(options.module)

    if found_suites is None:
        found_suites = find_suites(options, accept=module_accept)
    for suite in found_suites:
        for test, layer_name in tests_from_suite(suite, options,
                                                 accept=test_accept,
                                                 duplicated_test_ids=dupe_ids):
            if dupe_ids:
                # If there are any duplicated test IDs, we stop trying to
                # load tests; we'll raise an error later on with all the
                # duplicates in it.
                continue
            suite = suites.get(layer_name)
            if not suite:
                suite = suites[layer_name] = unittest.TestSuite()
            suite.addTest(test)
    if dupe_ids:
        message_lines = ['Duplicate test IDs found:'] + sorted(dupe_ids)
        message = '\n  '.join(message_lines)
        raise DuplicateTestIDError(message)
    return suites


def find_suites(options, accept=None):
    for fpath, package in find_test_files(options):
        for (prefix, prefix_package) in options.prefix:
            if fpath.startswith(prefix) and package == prefix_package:
                # strip prefix, strip .py suffix and convert separator to dots
                noprefix = fpath[len(prefix):]
                noext = strip_py_ext(options, noprefix)
                assert noext is not None
                module_name = noext.replace(os.path.sep, '.')
                if package:
                    module_name = package + '.' + module_name

                if accept is not None and not accept(module_name):
                    continue

                try:
                    module = import_name(module_name)
                except KeyboardInterrupt:
                    raise
                except BaseException:
                    exc_info = sys.exc_info()
                    if not options.post_mortem:
                        # Skip a couple of frames
                        exc_info = (
                            exc_info[:2] + (exc_info[2].tb_next.tb_next,))
                    suite = StartUpFailure(
                        options, module_name, exc_info)
                else:
                    try:
                        if hasattr(module, options.suite_name):
                            suite = getattr(module, options.suite_name)()
                        else:
                            loader = unittest.defaultTestLoader
                            suite = loader.loadTestsFromModule(module)
                            if suite.countTestCases() == 0:
                                raise TypeError(
                                    "Module %s does not define any tests"
                                    % module_name)

                        if not isinstance(suite, unittest.TestSuite):
                            # We extract the error message expression into a
                            # local variable because we want the `raise`
                            # statement to fit on a single line, to make the
                            # testrunner-debugging-import-failure.rst doctest
                            # see the same pdb output on Python 3.8 as on older
                            # Python versions.
                            bad_test_suite_msg = (
                                "Invalid test_suite, %r, in %s"
                                % (suite, module_name)
                            )
                            raise TypeError(bad_test_suite_msg)
                    except KeyboardInterrupt:
                        raise
                    except BaseException:
                        exc_info = sys.exc_info()
                        if not options.post_mortem:
                            # Suppress traceback
                            exc_info = exc_info[:2] + (None,)
                        suite = StartUpFailure(
                            options, module_name, exc_info)

                yield suite
                break


def find_test_files(options):
    found = {}
    for f, package in find_test_files_(options):
        if f not in found:
            found[f] = 1
            yield f, package


def find_test_files_(options):
    tests_pattern = options.tests_pattern
    test_file_pattern = options.test_file_pattern

    # If options.usecompiled, we can accept .pyc or .pyo files instead
    # of .py files.  We'd rather use a .py file if one exists.  `root2ext`
    # maps a test file path, sans extension, to the path with the best
    # extension found (.py if it exists, else .pyc or .pyo).
    # Note that "py" < "pyc" < "pyo", so if more than one extension is
    # found, the lexicographically smaller one is best.

    # Found a new test file, in directory `dirname`.  `noext` is the
    # file name without an extension, and `withext` is the file name
    # with its extension.
    def update_root2ext(dirname, noext, withext):
        key = os.path.join(dirname, noext)
        new = os.path.join(dirname, withext)
        if key in root2ext:
            root2ext[key] = min(root2ext[key], new)
        else:
            root2ext[key] = new

    for (p, package) in test_dirs(options, {}):
        for dirname, dirs, files in walk_with_symlinks(options, p):
            root2ext = {}
            dirs[:] = [
                d for d in dirs if identifier(d) and d not in IGNORE_FOLDERS
            ]
            d = os.path.split(dirname)[1]
            if tests_pattern(d) and contains_init_py(options, files):
                # tests directory
                for file in files:
                    noext = strip_py_ext(options, file)
                    if noext and test_file_pattern(noext):
                        update_root2ext(dirname, noext, file)

            for file in files:
                noext = strip_py_ext(options, file)
                if noext and tests_pattern(noext):
                    update_root2ext(dirname, noext, file)

            winners = sorted(root2ext.values())
            for file in winners:
                yield file, package


def strip_py_ext(options, path):
    """Return path without its .py (or .pyc or .pyo) extension, or None.

    If options.usecompiled is false:
        If path ends with ".py", the path without the extension is returned.
        Else None is returned.

    If options.usecompiled is true:
        If Python is running with -O, a .pyo extension is also accepted.
        If Python is running without -O, a .pyc extension is also accepted.
    """
    if path.endswith(".py"):
        return path[:-3]
    if options.usecompiled:
        if __debug__:
            # Python is running without -O.
            ext = ".pyc"
        else:
            # Python is running with -O.
            ext = ".pyo"
        if path.endswith(ext):
            return path[:-len(ext)]
    return None


def test_dirs(options, seen):
    if options.package:
        for p in options.package:
            p = import_name(p)
            for p in p.__path__:
                p = os.path.abspath(p)
                if p in seen:
                    continue
                for (prefix, package) in options.prefix:
                    if p.startswith(prefix) or p == prefix[:-1]:
                        seen[p] = 1
                        yield p, package
                        break
    else:
        yield from options.test_path


def walk_with_symlinks(options, dir):
    # TODO -- really should have test of this that uses symlinks
    #         this is hard on a number of levels ...
    for dirpath, dirs, files in os.walk(dir):
        dirs.sort()
        files.sort()
        dirs[:] = [d for d in dirs if d not in options.ignore_dir]
        yield (dirpath, dirs, files)
        for d in dirs:
            p = os.path.join(dirpath, d)
            if os.path.islink(p):
                yield from walk_with_symlinks(options, p)


compiled_suffixes = '.pyc', '.pyo'


def remove_stale_bytecode(options):
    if options.keepbytecode:
        return
    for (p, _) in options.test_path:
        for dirname, dirs, files in walk_with_symlinks(options, p):
            if '__pycache__' in dirs:
                # Do not recurse in there: we would end up removing all pyc
                # files because the main loop checks for py files in the same
                # directory.  Besides, stale pyc files in __pycache__ are
                # harmless, see PEP-3147 for details (sourceless imports
                # work only when pyc lives in the source dir directly).
                dirs.remove('__pycache__')

            for file in files:
                if file[-4:] in compiled_suffixes and file[:-1] not in files:
                    fullname = os.path.join(dirname, file)
                    options.output.info("Removing stale bytecode file %s"
                                        % fullname)
                    os.unlink(fullname)


def contains_init_py(options, fnamelist):
    """Return true iff fnamelist contains a suitable spelling of __init__.py.

    If options.usecompiled is false, this is so iff "__init__.py" is in
    the list.

    If options.usecompiled is true, then "__init__.pyo" is also acceptable
    if Python is running with -O, and "__init__.pyc" is also acceptable if
    Python is running without -O.
    """
    if "__init__.py" in fnamelist:
        return True
    if options.usecompiled:
        if __debug__:
            # Python is running without -O.
            return "__init__.pyc" in fnamelist
        else:
            # Python is running with -O.
            return "__init__.pyo" in fnamelist
    return False


def import_name(name):
    __import__(name)
    return sys.modules[name]


def tests_from_suite(suite, options, dlevel=1,
                     dlayer=zope.testrunner.layer.UnitTests,
                     accept=None,
                     seen_test_ids=None, duplicated_test_ids=None):
    """Returns a sequence of (test, layer_name)

    The tree of suites is recursively visited, with the most specific
    layer taking precedence. So if a TestCase with a layer of 'foo' is
    contained in a TestSuite with a layer of 'bar', the test case would be
    returned with 'foo' as the layer.

    Tests are also filtered out based on the test level and accept predicate.
    accept is a function, returning boolean for given test name (see also
    build_filtering_func()).
    """
    # We use this to track the test IDs that have been registered.
    # tests_from_suite will complain if it encounters the same test ID
    # twice.
    if seen_test_ids is None:
        seen_test_ids = set()
    if duplicated_test_ids is None:
        duplicated_test_ids = set()

    level = getattr(suite, 'level', dlevel)
    layer = getattr(suite, 'layer', dlayer)
    if not isinstance(layer, str):
        layer = name_from_layer(layer)

    if isinstance(suite, unittest.TestSuite):
        for possible_suite in suite:
            yield from tests_from_suite(
                possible_suite, options, level, layer,
                accept=accept,
                seen_test_ids=seen_test_ids,
                duplicated_test_ids=duplicated_test_ids)
    elif isinstance(suite, StartUpFailure):
        yield (suite, None)
    else:
        if options.require_unique_ids:
            suite_id = str(suite)
            if suite_id in seen_test_ids:
                duplicated_test_ids.add(suite_id)
            else:
                seen_test_ids.add(suite_id)
        if options.only_level is None:
            if options.at_level <= 0 or level <= options.at_level:
                if accept is None or accept(str(suite)):
                    yield (suite, layer)
        else:
            if level == options.only_level:
                if accept is None or accept(str(suite)):
                    yield (suite, layer)


_layer_name_cache = {}


def name_from_layer(layer):
    """Determine a name for the Layer using the namespace to avoid conflicts.

    We also cache a name -> layer mapping to enable layer_from_name to work
    in cases where the layer cannot be imported (such as layers defined
    in doctests)
    """
    if layer.__module__ == '__builtin__':
        name = layer.__name__
    else:
        name = layer.__module__ + '.' + layer.__name__
    _layer_name_cache[name] = layer
    return name


class Find(zope.testrunner.feature.Feature):
    """Finds tests and registers them with the test runner."""

    active = True

    def global_setup(self):
        # Add directories to the path
        for path in reversed(self.runner.options.path):
            if path not in sys.path:
                sys.path.insert(0, path)

        tests = find_tests(self.runner.options, self.runner.found_suites)
        self.import_errors = tests.pop(None, None)
        self.runner.register_tests(tests)

        # XXX move to reporting ???
        self.runner.options.output.import_errors(self.import_errors)
        self.runner.import_errors = list(self.import_errors or [])

    def report(self):
        self.runner.options.output.modules_with_import_problems(
            self.import_errors)



##############################################################################
#
# Copyright (c) 2013 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

if __name__ == '__main__':
    # allow people to try out the test runner with
    # python -m zope.testrunner --test-path .
    from zope.testrunner import run
    run()



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import doctest


pledge_template = """\
I give my pledge, as %s,
to save, and faithfully, to defend from waste,
the natural resources of my %s.
It's soils, minerals, forests, waters, and wildlife.
"""


def pledge():
    """
    >>> def print_pledge():
    ...     print(pledge_template % ('and earthling', 'planet'))
    >>> print_pledge()
    I give my pledge, as an earthling,
    to save, and faithfully, to defend from waste,
    the natural resources of my planet.
    It's soils, minerals, forests, waters, and wildlife.
    <BLANKLINE>
    """


def test_suite():
    return doctest.DocTestSuite()



##############################################################################
#
# Copyright (c) 2004-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Logging support.

This code is pretty much untested and was only mechanically refactored.

The module name is not 'logging' because of a name collision with Python's
logging module.
"""
import logging
import logging.config
import os
import os.path
import warnings

import zope.testrunner.feature


class Logging(zope.testrunner.feature.Feature):

    active = True

    def global_setup(self):
        # Get the log.ini file either from the envvar
        # ``ZOPE_TESTRUNNER_LOG_INI`` or file ``log.ini`` in
        # the current working directory.
        logini = os.environ.get("ZOPE_TESTRUNNER_LOG_INI")
        if logini is not None and not os.path.exists(logini):
            warnings.warn(
                "ERROR: file specified by envvar ZOPE_TESTRUNNER_LOG_INI` "
                "does not exist")
            logini = None
        if logini is None:
            logini = "log.ini"
        logini = os.path.abspath(logini)  # make absolute
        if os.path.exists(logini):
            logging.config.fileConfig(logini, disable_existing_loggers=False)
            # remember the log configuration in envvar for use
            # by child processes
            os.environ["ZOPE_TESTRUNNER_LOG_INI"] = logini
        else:
            # If there's no log.ini, cause the logging package to be
            # silent during testing.
            root = logging.getLogger()
            root.addHandler(NullHandler())
            logging.basicConfig()

        if "LOGGING" in os.environ:
            level = int(os.environ["LOGGING"])
            logging.getLogger().setLevel(level)


class NullHandler(logging.Handler):
    """Logging handler that drops everything on the floor.

    We require silence in the test environment.  Hush.
    """

    def emit(self, record):
        pass



from logging import DEBUG
from logging import getLogger
from os import chdir
from os import environ
from os import getcwd
from os.path import dirname
from os.path import join
from unittest import TestCase
from warnings import catch_warnings

from ..logsupport import Logging


setup_logging = Logging.global_setup
setup_logging = getattr(setup_logging, "__func__", setup_logging)
evn = "ZOPE_TESTRUNNER_LOG_INI"


class LogsupportTests(TestCase):
    def setUp(self):
        # save root logger config
        logger = getLogger()
        self.logconfig = lc = logger.__dict__.copy()
        for n in ("filters", "handlers"):
            lc[n] = lc[n][:]  # copy
            del getattr(logger, n)[:]  # clear
        # save working directory
        self.cwd = getcwd()
        # save and clear envvar
        self.envvar = ev = environ.get(evn, self)
        if ev is not self:
            del environ[evn]

    def tearDown(self):
        # restore working directory
        chdir(self.cwd)
        # restore root logger configuration
        logger = getLogger()
        logger.__dict__.update(self.logconfig)
        # restore envvar
        ev = self.envvar
        if ev is not self:
            environ[evn] = ev
        elif evn in environ:
            del environ[evn]

    def test_via_cwd(self):
        chdir(join(dirname(__file__), "logsupport"))
        self.check_logging()

    def test_via_envvar(self):
        lc = join(dirname(__file__), "logsupport", "log.ini")
        environ[evn] = lc
        self.check_logging()

    def test_via_bad_envvar(self):
        chdir(join(dirname(__file__), "logsupport"))
        environ[evn] = "not_existing"
        with catch_warnings(record=True) as w:
            self.check_logging()
            self.assertEqual(len(w), 1)

    def check_logging(self):
        setup_logging(None)
        logger = getLogger()
        self.assertEqual(logger.level, DEBUG)
        self.assertEqual(len(logger.handlers), 1)
        lc = join(dirname(__file__), "logsupport", "log.ini")
        self.assertEqual(environ[evn], lc)



##############################################################################
#
# Copyright (c) 2004-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Doc test support for the test runner.
"""

import doctest
import sys

import zope.testrunner.feature
from zope.testrunner.exceptions import DocTestFailureException


class DocTest(zope.testrunner.feature.Feature):

    active = True

    def global_setup(self):
        options = self.runner.options
        output = options.output

        self.old_reporting_flags = doctest.set_unittest_reportflags(0)

        reporting_flags = 0
        if options.ndiff:
            reporting_flags = doctest.REPORT_NDIFF
        if options.udiff:
            if reporting_flags:
                output.error(
                    "Can only give one of --ndiff, --udiff, or --cdiff")
                sys.exit(1)
            reporting_flags = doctest.REPORT_UDIFF
        if options.cdiff:
            if reporting_flags:
                output.error(
                    "Can only give one of --ndiff, --udiff, or --cdiff")
                sys.exit(1)
            reporting_flags = doctest.REPORT_CDIFF
        if options.report_only_first_failure:
            reporting_flags |= doctest.REPORT_ONLY_FIRST_FAILURE

        if reporting_flags:
            doctest.set_unittest_reportflags(reporting_flags)

    def global_shutdown(self):
        doctest.set_unittest_reportflags(self.old_reporting_flags)


# Use a special exception for the test runner.
doctest.DocTestCase.failureException = DocTestFailureException



##############################################################################
#
# Copyright (c) 2001, 2002 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Set up testing environment
"""

import sys
import traceback

import zope.exceptions.exceptionformatter

import zope.testrunner.feature


try:
    from traceback import _parse_value_tb
    from traceback import _sentinel
except ImportError:
    # before 3.10
    # for Python before 3.10, the first 3 parameters of
    # ``print_exception`` and ``format_exception`` are all mandatory
    # and the first one (``etype`` alias ``t``) is ignored
    _sentinel = object()

    def _parse_value_tb(ignored, value, tb):
        return value, tb


def _iter_chain(exc, custom_tb=None, seen=None):
    if seen is None:
        seen = set()
    seen.add(exc)
    its = []
    context = exc.__context__
    cause = exc.__cause__
    if cause is not None and cause not in seen:
        its.append(_iter_chain(cause, False, seen))
        its.append([(traceback._cause_message, None)])
    elif (context is not None and
            not exc.__suppress_context__ and
            context not in seen):
        its.append(_iter_chain(context, None, seen))
        its.append([(traceback._context_message, None)])
    its.append([(exc, custom_tb or exc.__traceback__)])
    # itertools.chain is in an extension module and may be unavailable
    for it in its:
        yield from it


def format_exception(t, value=_sentinel, tb=_sentinel, limit=None, chain=None):
    v, tb = _parse_value_tb(t, value, tb)
    if chain:
        values = _iter_chain(v, tb)
    else:
        values = [(v, tb)]
    fmt = zope.exceptions.exceptionformatter.TextExceptionFormatter(
        limit=None, with_filenames=True)
    for v, tb in values:
        if isinstance(v, str):
            return v
        return fmt.formatException(t, v, tb)


def print_exception(t, value=_sentinel, tb=_sentinel,
                    limit=None, file=None, chain=None):
    v, tb = _parse_value_tb(t, value, tb)
    if chain:
        values = _iter_chain(v, tb)
    else:
        values = [(v, tb)]
    if file is None:
        file = sys.stdout
    for v, tb in values:
        file.writelines(format_exception(t, v, tb, limit))


class Traceback(zope.testrunner.feature.Feature):

    active = True

    def global_setup(self):
        self.old_format = traceback.format_exception
        traceback.format_exception = format_exception

        self.old_print = traceback.print_exception
        traceback.print_exception = print_exception

    def global_teardown(self):
        traceback.format_exception = self.old_format
        traceback.print_exception = self.old_print



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Sample package that knits in extra directories.
"""

import os


__path__.append(
    os.path.join(
        os.path.dirname(  # testing
            os.path.dirname(  # testrunner-ex-knit-lib
                os.path.dirname(  # sample4
                    os.path.dirname(__file__)  # products
                )
            )
        ),
        "testrunner-ex-pp-products"
    )
)



##############################################################################
#
# Copyright (c) 2010 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unit tests for the testrunner's subunit integration."""


import io
import sys
import unittest

from zope.testrunner import formatter


try:
    import subunit
    subunit
except ImportError:
    def test_suite():
        return unittest.TestSuite()
else:

    class TestSubunitTracebackPrintingMixin:

        def makeByteStringFailure(self, text, encoding):
            try:
                # It's more accurate to just use the text directly.
                self.fail(text)
            except self.failureException:
                return sys.exc_info()

        def test_print_failure_containing_utf8_bytestrings(self):
            exc_info = self.makeByteStringFailure(chr(6514), 'utf8')
            self.subunit_formatter.test_failure(self, 0, exc_info)
            assert b"AssertionError: \xe1\xa5\xb2" in self.output.getvalue()
            # '\xe1\xa5\xb2'.decode('utf-8') == chr(6514)

        def test_print_error_containing_utf8_bytestrings(self):
            exc_info = self.makeByteStringFailure(chr(6514), 'utf8')
            self.subunit_formatter.test_error(self, 0, exc_info)
            assert b"AssertionError: \xe1\xa5\xb2" in self.output.getvalue()
            # '\xe1\xa5\xb2'.decode('utf-8') == chr(6514)

    class TestSubunitTracebackPrinting(
            TestSubunitTracebackPrintingMixin, unittest.TestCase):

        def setUp(self):
            class FormatterOptions:
                verbose = False
            options = FormatterOptions()

            self.output = io.BytesIO()
            self.subunit_formatter = formatter.SubunitOutputFormatter(
                options, stream=self.output)

    class TestSubunitV2TracebackPrinting(
            TestSubunitTracebackPrintingMixin, unittest.TestCase):

        def setUp(self):
            class FormatterOptions:
                verbose = False
            options = FormatterOptions()

            self.output = io.BytesIO()
            self.subunit_formatter = formatter.SubunitV2OutputFormatter(
                options, stream=self.output)

    def test_suite():
        return unittest.defaultTestLoader.loadTestsFromName(__name__)



##############################################################################
#
# Copyright (c) 2022 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from _thread import start_new_thread
from threading import Lock
from threading import Thread
from time import sleep
from unittest import TestCase
from unittest import skipUnless

from ..threadsupport import current_frames
from ..threadsupport import enumerate


class ThreadMixin:
    """test thread."""

    def __init__(self):
        self.lock = Lock()
        self.stopped = False

    def run(self):
        with self.lock:
            pass
        self.stopped = True


class ThrThread(ThreadMixin, Thread):
    """``threading.Thread`` thread."""

    def __init__(self, name):
        Thread.__init__(self, name=name)
        ThreadMixin.__init__(self)


class DummyThread(ThreadMixin):
    def start(self):
        start_new_thread(self.run, ())


class Tests(TestCase):
    def setUp(self):
        self._threads = []
        self._prethreads = enumerate()

    def tearDown(self):
        for t in self._threads:
            if not t.stopped:
                t.lock.release()
                sleep(0.01)

    def _mk_thread(self, name=None):
        """create a new thread and start it.

        Use a full thread if *name*; otherwise a dummy thread.
        """
        t = ThrThread(name) if name else DummyThread()
        t.lock.acquire()
        t.start()
        self._threads.append(t)
        sleep(0.01)
        return t

    def alive(self):
        """alive threads not in ``_prethreads``."""
        return [p for p in enumerate()
                if p.is_alive() and p not in self._prethreads]

    def test_thr_proxy(self):
        self.check_proxy("Test ThrThread")

    @skipUnless(current_frames, "no `current_frames`")
    def test_dummy_proxy(self):
        self.check_proxy()

    def check_proxy(self, name=None):
        t = self._mk_thread(name)
        prs = self.alive()
        self.assertEqual(len(prs), 1)
        pr = prs[0]
        self.assertTrue(pr.is_alive())
        if name is None:
            self.assertTrue(pr.name.startswith("Dummy-"))
            self.assertIn("DummyThread", repr(pr))
        else:
            self.assertEqual(pr.name, t.name)
            self.assertEqual(repr(pr), repr(t))

    def test_thr_new_thread(self):
        self.check_new_thread("Test ThrThread")

    @skipUnless(current_frames, "no `current_frames`")
    def test_dummy_new_thread(self):
        self.check_new_thread()

    def check_new_thread(self, name=None):
        self._mk_thread(name and (name + "-1"))
        threads = self.alive()
        self._mk_thread(name and (name + "-2"))
        new_threads = [p for p in self.alive() if p not in threads]
        self.assertEqual(len(new_threads), 1)

    def test_thr_stopped(self):
        self.check_stopped("Test ThrThread")

    @skipUnless(current_frames, "no `current_frames`")
    def test_dummy_stopped(self):
        self.check_stopped()

    def check_stopped(self, name=None):
        t = self._mk_thread(name)
        t.lock.release()
        sleep(0.01)
        self.assertEqual(self.alive(), [])



##############################################################################
#
# Copyright (c) 2003 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

import doctest
import unittest


def f():
    g()


def g():
    x = 1
    x = x + 1
    x = y + 1  # noqa: F821
    x = x + 1


def eek(self):
    """
    >>> f()
    1
    """


class Test(unittest.TestCase):

    def test1(self):
        pass

    def test2(self):
        pass

    def test3(self):
        f()

    def test4(self):
        pass

    def test5(self):
        pass


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(doctest.DocTestSuite())
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(Test))
    suite.addTest(doctest.DocFileSuite('e.rst'))
    return suite



##############################################################################
#
# Copyright (c) 2004-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Self-test support.

Provides setup routines that enable the test runner to test itself.
"""

import pdb  # noqa: T100 import for pdb found

import zope.testrunner.feature


real_pdb_set_trace = pdb.set_trace


class SelfTest(zope.testrunner.feature.Feature):

    active = True

    def global_setup(self):
        # Make sure we start with real pdb.set_trace.
        pdb.set_trace = real_pdb_set_trace



##############################################################################
#
# Copyright (c) 2004-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Output formatting.
"""
import doctest
import io
import os
import re
import socket
import sys
import tempfile
import traceback
from collections.abc import MutableMapping
from contextlib import contextmanager
from contextlib import suppress
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from datetime import timedelta
from pathlib import Path
from xml.etree import ElementTree

from zope.testrunner.exceptions import DocTestFailureException
from zope.testrunner.find import StartUpFailure


try:
    import manuel.testing
    HAVE_MANUEL = True
except ImportError:
    HAVE_MANUEL = False


doctest_template = """
File "%s", line %s, in %s

%s
Want:
%s
Got:
%s
"""


class OutputFormatter:
    """Test runner output formatter."""

    # Implementation note: be careful about printing stuff to sys.stderr.
    # It is used for interprocess communication between the parent and the
    # child test runner, when you run some test layers in a subprocess.
    # resume_layer() reasigns sys.stderr for this reason, but be careful
    # and don't store the original one in __init__ or something.

    max_width = 80

    def __init__(self, options):
        self.options = options
        self.last_width = 0
        self.compute_max_width()

    progress = property(lambda self: self.options.progress)
    verbose = property(lambda self: self.options.verbose)
    in_subprocess = property(
        lambda self: (
            self.options.resume_layer is not None and
            self.options.processes > 1))

    def compute_max_width(self):
        """Try to determine the terminal width."""
        # Note that doing this every time is more test friendly.
        self.max_width = tigetnum('cols', self.max_width)

    def getShortDescription(self, test, room):
        """Return a description of a test that fits in ``room`` characters."""
        room -= 1
        s = str(test)
        if len(s) > room:
            pos = s.find(" (")
            if pos >= 0:
                w = room - (pos + 5)
                if w < 1:
                    # first portion (test method name) is too long
                    s = s[:room - 3] + "..."
                else:
                    pre = s[:pos + 2]
                    post = s[-w:]
                    s = f"{pre}...{post}"
            else:
                w = room - 4
                s = '... ' + s[-w:]

        return ' ' + s[:room]

    def info(self, message):
        """Print an informative message."""
        print(message)

    def info_suboptimal(self, message):
        """Print an informative message about losing some of the features.

        For example, when you run some tests in a subprocess, you lose the
        ability to use the debugger.
        """
        print(message)

    def error(self, message):
        """Report an error."""
        print(message)

    def error_with_banner(self, message):
        """Report an error with a big ASCII banner."""
        print()
        print('*' * 70)
        self.error(message)
        print('*' * 70)
        print()

    def profiler_stats(self, stats):
        """Report profiler stats."""
        stats.print_stats(50)

    def import_errors(self, import_errors):
        """Report test-module import errors (if any)."""
        if import_errors:
            print("Test-module import failures:")
            for error in import_errors:
                self.print_traceback("Module: %s\n" % error.module,
                                     error.exc_info),
            print()

    def tests_with_errors(self, errors):
        """Report names of tests with errors (if any)."""
        if errors:
            print()
            print("Tests with errors:")
            for test, exc_info in errors:
                print("  ", test)

    def tests_with_failures(self, failures):
        """Report names of tests with failures (if any)."""
        if failures:
            print()
            print("Tests with failures:")
            for test, exc_info in failures:
                print("  ", test)

    def modules_with_import_problems(self, import_errors):
        """Report names of modules with import problems (if any)."""
        if import_errors:
            print()
            print("Test-modules with import problems:")
            for test in import_errors:
                print("  " + test.module)

    def format_seconds(self, n_seconds):
        """Format a time in seconds."""
        if n_seconds >= 60:
            n_minutes, n_seconds = divmod(n_seconds, 60)
            return "%d minutes %.3f seconds" % (n_minutes, n_seconds)
        else:
            return "%.3f seconds" % n_seconds

    def format_seconds_short(self, n_seconds):
        """Format a time in seconds (short version)."""
        return "%.3f s" % n_seconds

    def summary(self, n_tests, n_failures, n_errors, n_seconds,
                n_skipped=0):
        """Summarize the results of a single test layer."""
        print("  Ran %s tests with %s failures, %s errors and "
              "%s skipped in %s."
              % (n_tests, n_failures, n_errors, n_skipped,
                  self.format_seconds(n_seconds)))

    def totals(self, n_tests, n_failures, n_errors, n_seconds,
               n_skipped=0):
        """Summarize the results of all layers."""
        print("Total: %s tests, %s failures, %s errors and %s skipped in %s."
              % (n_tests, n_failures, n_errors, n_skipped,
                 self.format_seconds(n_seconds)))

    def list_of_tests(self, tests, layer_name):
        """Report a list of test names."""
        print("Listing %s tests:" % layer_name)
        for test in tests:
            print(' ', test)

    def garbage(self, garbage):
        """Report garbage generated by tests."""
        if garbage:
            print("Tests generated new (%d) garbage:" % len(garbage))
            print(garbage)

    def test_garbage(self, test, garbage):
        """Report garbage generated by a test."""
        if garbage:
            print("The following test left garbage:")
            print(test)
            print(garbage)

    def test_threads(self, test, new_threads):
        """Report threads left behind by a test."""
        if new_threads:
            print("The following test left new threads behind:")
            print(test)
            print("New thread(s):", new_threads)

    def test_cycles(self, test, cycles):
        """Report cyclic garbage left behind by a test."""
        if cycles:
            print("The following test left cyclic garbage behind:")
            print(test)
            for i, cy in enumerate(cycles):
                print("Cycle", i + 1)
                for oi in cy:
                    print(" * ", "\n   ".join(oi))

    def refcounts(self, rc, prev):
        """Report a change in reference counts."""
        print("  sys refcount=%-8d change=%-6d" % (rc, rc - prev))

    def detailed_refcounts(self, track, rc, prev):
        """Report a change in reference counts, with extra detail."""
        print("  sum detail refcount=%-8d"
              " sys refcount=%-8d"
              " change=%-6d"
              % (track.n, rc, rc - prev))
        track.output()

    def start_set_up(self, layer_name):
        """Report that we're setting up a layer.

        The next output operation should be stop_set_up().
        """
        print("  Set up %s" % layer_name, end=' ')
        sys.stdout.flush()

    def stop_set_up(self, seconds):
        """Report that we've set up a layer.

        Should be called right after start_set_up().
        """
        print("in %s." % self.format_seconds(seconds))

    def start_tear_down(self, layer_name):
        """Report that we're tearing down a layer.

        The next output operation should be stop_tear_down() or
        tear_down_not_supported().
        """
        print("  Tear down %s" % layer_name, end=' ')
        sys.stdout.flush()

    def stop_tear_down(self, seconds):
        """Report that we've tore down a layer.

        Should be called right after start_tear_down().
        """
        print("in %s." % self.format_seconds(seconds))

    def tear_down_not_supported(self):
        """Report that we could not tear down a layer.

        Should be called right after start_tear_down().
        """
        print("... not supported")

    def start_test(self, test, tests_run, total_tests):
        """Report that we're about to run a test.

        The next output operation should be test_success(), test_error(), or
        test_failure().
        """
        self.test_width = 0
        if self.progress:
            if self.last_width:
                sys.stdout.write('\r' + (' ' * self.last_width) + '\r')

            s = "    %d/%d (%.1f%%)" % (tests_run, total_tests,
                                        tests_run * 100.0 / total_tests)
            sys.stdout.write(s)
            self.test_width += len(s)
            if self.verbose == 1:
                room = self.max_width - self.test_width - 1
                s = self.getShortDescription(test, room)
                sys.stdout.write(s)
                self.test_width += len(s)

        elif self.verbose == 1:
            sys.stdout.write('.' * test.countTestCases())

        elif self.in_subprocess:
            sys.stdout.write('.' * test.countTestCases())
            # Give the parent process a new line so it sees the progress
            # in a timely manner.
            sys.stdout.write('\n')

        if self.verbose > 1:
            s = str(test)
            sys.stdout.write(' ')
            sys.stdout.write(s)
            self.test_width += len(s) + 1

        sys.stdout.flush()

    def test_success(self, test, seconds):
        """Report that a test was successful.

        Should be called right after start_test().

        The next output operation should be stop_test().
        """
        if self.verbose > 2:
            s = " (%s)" % self.format_seconds_short(seconds)
            sys.stdout.write(s)
            self.test_width += len(s) + 1

    def test_skipped(self, test, reason):
        """Report that a test was skipped.

        Should be called right after start_test().

        The next output operation should be stop_test().
        """
        if self.verbose > 2:
            s = " (skipped: %s)" % reason
        elif self.verbose > 1:
            s = " (skipped)"
        else:
            return
        sys.stdout.write(s)
        self.test_width += len(s) + 1

    def test_error(self, test, seconds, exc_info, stdout=None, stderr=None):
        """Report that an error occurred while running a test.

        Should be called right after start_test().

        The next output operation should be stop_test().
        """
        if self.verbose > 2:
            print(" (%s)" % self.format_seconds_short(seconds))
        print()
        self.print_traceback("Error in test %s" % test, exc_info)
        self.print_std_streams(stdout, stderr)
        self.test_width = self.last_width = 0

    def test_failure(self, test, seconds, exc_info, stdout=None, stderr=None):
        """Report that a test failed.

        Should be called right after start_test().

        The next output operation should be stop_test().
        """
        if self.verbose > 2:
            print(" (%s)" % self.format_seconds_short(seconds))
        print()
        self.print_traceback("Failure in test %s" % test, exc_info)
        self.print_std_streams(stdout, stderr)
        self.test_width = self.last_width = 0

    def print_traceback(self, msg, exc_info):
        """Report an error with a traceback."""
        print()
        print(msg)
        print(self.format_traceback(exc_info))

    def print_std_streams(self, stdout, stderr):
        """Emit contents of buffered standard streams."""
        if stdout:
            sys.stdout.write("Stdout:\n")
            sys.stdout.write(stdout)
            if not stdout.endswith("\n"):
                sys.stdout.write("\n")
            sys.stdout.write("\n")
        if stderr:
            sys.stderr.write("Stderr:\n")
            sys.stderr.write(stderr)
            if not stderr.endswith("\n"):
                sys.stderr.write("\n")
            sys.stderr.write("\n")

    def format_traceback(self, exc_info):
        """Format the traceback."""
        v = exc_info[1]
        if isinstance(v, DocTestFailureException):
            tb = v.args[0]
        elif isinstance(v, doctest.DocTestFailure):
            tb = doctest_template % (
                v.test.filename,
                v.test.lineno + v.example.lineno + 1,
                v.test.name,
                v.example.source,
                v.example.want,
                v.got,
            )
        else:
            tb = "".join(traceback.format_exception(*exc_info))
        return tb

    def stop_test(self, test, gccount):
        """Clean up the output state after a test."""
        if gccount and self.verbose:
            s = "!" if self.verbose == 1 else " [%d]" % gccount
            self.test_width += len(s)
            print(s, end="")
        if self.progress:
            self.last_width = self.test_width
        elif self.verbose > 1:
            print()
        sys.stdout.flush()

    def stop_tests(self):
        """Clean up the output state after a collection of tests."""
        if self.progress and self.last_width:
            sys.stdout.write('\r' + (' ' * self.last_width) + '\r')
        if self.verbose == 1 or self.progress:
            print()


def tigetnum(attr, default=None):
    """Return a value from the terminfo database.

    Terminfo is used on Unix-like systems to report various terminal attributes
    (such as width, height or the number of supported colors).

    Returns ``default`` when the ``curses`` module is not available, or when
    sys.stdout is not a terminal.
    """
    try:
        import curses
    except ImportError:
        # avoid reimporting a broken module
        sys.modules['curses'] = None
    else:
        # If sys.stdout is not a real file object (e.g. in unit tests that
        # use various wrappers), you get an error, different depending on
        # Python version:
        expected_exceptions = (
            curses.error, TypeError, AttributeError, io.UnsupportedOperation)
        try:
            curses.setupterm()
        except expected_exceptions:
            # You get curses.error when $TERM is set to an unknown name
            pass
        else:
            try:
                return curses.tigetnum(attr)
            except expected_exceptions:
                # You get TypeError on PyPy3 due to a bug:
                # https://bitbucket.org/pypy/pypy/issue/2016/pypy3-cursestigetnum-raises-ctype
                pass
    return default


def terminal_has_colors():
    """Determine whether the terminal supports colors.

    Some terminals (e.g. the emacs built-in one) don't.
    """
    return tigetnum('colors', -1) >= 8


class ColorfulOutputFormatter(OutputFormatter):
    """Output formatter that uses ANSI color codes.

    Like syntax highlighting in your text editor, colorizing
    test failures helps the developer.
    """

    # These colors are carefully chosen to have enough contrast
    # on terminals with both black and white background.
    colorscheme = {'normal': 'normal',
                   'default': 'default',
                   'info': 'normal',
                   'suboptimal-behaviour': 'magenta',
                   'error': 'brightred',
                   'number': 'green',
                   'slow-test': 'brightmagenta',
                   'ok-number': 'green',
                   'error-number': 'brightred',
                   'filename': 'lightblue',
                   'lineno': 'lightred',
                   'testname': 'lightcyan',
                   'failed-example': 'cyan',
                   'expected-output': 'green',
                   'actual-output': 'red',
                   'character-diffs': 'magenta',
                   'diff-chunk': 'magenta',
                   'exception': 'red',
                   'skipped': 'brightyellow',
                   }

    # Map prefix character to color in diff output.  This handles ndiff and
    # udiff correctly, but not cdiff.  In cdiff we ought to highlight '!' as
    # expected-output until we see a '-', then highlight '!' as actual-output,
    # until we see a '*', then switch back to highlighting '!' as
    # expected-output.  Nevertheless, coloried cdiffs are reasonably readable,
    # so I'm not going to fix this.
    #   -- mgedmin
    diff_color = {'-': 'expected-output',
                  '+': 'actual-output',
                  '?': 'character-diffs',
                  '@': 'diff-chunk',
                  '*': 'diff-chunk',
                  '!': 'actual-output',
                  }

    prefixes = [('dark', '0;'),
                ('light', '1;'),
                ('bright', '1;'),
                ('bold', '1;'),
                ]

    colorcodes = {'default': 0, 'normal': 0,
                  'black': 30,
                  'red': 31,
                  'green': 32,
                  'brown': 33, 'yellow': 33,
                  'blue': 34,
                  'magenta': 35,
                  'cyan': 36,
                  'grey': 37, 'gray': 37, 'white': 37}

    slow_test_threshold = 10.0  # seconds

    def color_code(self, color):
        """Convert a color description (e.g. 'lightred') to a terminal code."""
        prefix_code = ''
        for prefix, code in self.prefixes:
            if color.startswith(prefix):
                color = color[len(prefix):]
                prefix_code = code
                break
        color_code = self.colorcodes[color]
        return f'\033[{prefix_code}{color_code}m'

    def color(self, what):
        """Pick a named color from the color scheme"""
        return self.color_code(self.colorscheme[what])

    def colorize(self, what, message, normal='normal'):
        """Wrap message in color."""
        return self.color(what) + message + self.color(normal)

    def error_count_color(self, n):
        """Choose a color for the number of errors."""
        if n:
            return self.color('error-number')
        else:
            return self.color('ok-number')

    def skip_count_color(self, n):
        """Choose a color for the number of skipped tests."""
        if n:
            return self.color('skipped')
        else:
            return self.color('ok-number')

    def test_skipped(self, test, reason):
        """Report that a test was skipped.

        Should be called right after start_test().

        The next output operation should be stop_test().
        """
        if self.verbose > 2:
            s = " ({}skipped: {}{})".format(
                self.color('skipped'), reason, self.color('info'))
        elif self.verbose > 1:
            s = " ({}skipped{})".format(
                self.color('skipped'), self.color('info'))
        else:
            return
        sys.stdout.write(s)
        self.test_width += len(s) + 1

    def info(self, message):
        """Print an informative message."""
        print(self.colorize('info', message))

    def info_suboptimal(self, message):
        """Print an informative message about losing some of the features.

        For example, when you run some tests in a subprocess, you lose the
        ability to use the debugger.
        """
        print(self.colorize('suboptimal-behaviour', message))

    def error(self, message):
        """Report an error."""
        print(self.colorize('error', message))

    def error_with_banner(self, message):
        """Report an error with a big ASCII banner."""
        print()
        print(self.colorize('error', '*' * 70))
        self.error(message)
        print(self.colorize('error', '*' * 70))
        print()

    def tear_down_not_supported(self):
        """Report that we could not tear down a layer.

        Should be called right after start_tear_down().
        """
        print("...", self.colorize('suboptimal-behaviour', "not supported"))

    def format_seconds(self, n_seconds, normal='normal'):
        """Format a time in seconds."""
        if n_seconds >= 60:
            n_minutes, n_seconds = divmod(n_seconds, 60)
            return "{} minutes {} seconds".format(
                self.colorize('number', '%d' % n_minutes, normal),
                self.colorize('number', '%.3f' % n_seconds, normal))
        else:
            return "%s seconds" % (
                self.colorize('number', '%.3f' % n_seconds, normal))

    def format_seconds_short(self, n_seconds):
        """Format a time in seconds (short version)."""
        if n_seconds >= self.slow_test_threshold:
            color = 'slow-test'
        else:
            color = 'number'
        return self.colorize(color, "%.3f s" % n_seconds)

    def summary(self, n_tests, n_failures, n_errors, n_seconds,
                n_skipped=0):
        """Summarize the results."""
        sys.stdout.writelines([
            self.color('info'), '  Ran ',
            self.color('number'), str(n_tests),
            self.color('info'), ' tests with ',
            self.error_count_color(n_failures), str(n_failures),
            self.color('info'), ' failures, ',
            self.error_count_color(n_errors), str(n_errors),
            self.color('info'), ' errors, ',
            self.skip_count_color(n_skipped), str(n_skipped),
            self.color('info'), ' skipped in ',
            self.format_seconds(n_seconds, 'info'), '.',
            self.color('normal'), '\n',
        ])

    def totals(self, n_tests, n_failures, n_errors, n_seconds,
               n_skipped=0):
        """Report totals (number of tests, failures, and errors)."""
        sys.stdout.writelines([
            self.color('info'), 'Total: ',
            self.color('number'), str(n_tests),
            self.color('info'), ' tests, ',
            self.error_count_color(n_failures), str(n_failures),
            self.color('info'), ' failures, ',
            self.error_count_color(n_errors), str(n_errors),
            self.color('info'), ' errors, ',
            self.skip_count_color(n_skipped), str(n_skipped),
            self.color('info'), ' skipped in ',
            self.format_seconds(n_seconds, 'info'), '.',
            self.color('normal'), '\n'])

    def print_traceback(self, msg, exc_info):
        """Report an error with a traceback."""
        print()
        print(self.colorize('error', msg))
        v = exc_info[1]
        if isinstance(v, DocTestFailureException):
            self.print_doctest_failure(v.args[0])
        elif isinstance(v, doctest.DocTestFailure):
            # I don't think these are ever used... -- mgedmin
            tb = self.format_traceback(exc_info)
            print(tb)
        else:
            tb = self.format_traceback(exc_info)
            self.print_colorized_traceback(tb)

    def print_doctest_failure(self, formatted_failure):
        """Report a doctest failure.

        ``formatted_failure`` is a string -- that's what
        DocTestSuite/DocFileSuite gives us.
        """
        color_of_indented_text = 'normal'
        colorize_diff = False
        for line in formatted_failure.splitlines():
            if line.startswith('File '):
                m = re.match(r'File "(.*)", line (\d*), in (.*)$', line)
                if m:
                    filename, lineno, test = m.groups()
                    sys.stdout.writelines([
                        self.color('normal'), 'File "',
                        self.color('filename'), filename,
                        self.color('normal'), '", line ',
                        self.color('lineno'), lineno,
                        self.color('normal'), ', in ',
                        self.color('testname'), test,
                        self.color('normal'), '\n'])
                else:
                    print(line)
            elif line.startswith('    ') or line.strip() == '':
                if colorize_diff and len(line) > 4:
                    color = self.diff_color.get(
                        line[4], color_of_indented_text)
                    print(self.colorize(color, line))
                else:
                    if line.strip() != '':
                        print(self.colorize(color_of_indented_text, line))
                    else:
                        print(line)
            else:
                colorize_diff = False
                if line.startswith('Failed example'):
                    color_of_indented_text = 'failed-example'
                elif line.startswith('Expected:'):
                    color_of_indented_text = 'expected-output'
                elif line.startswith('Got:'):
                    color_of_indented_text = 'actual-output'
                elif line.startswith('Exception raised:'):
                    color_of_indented_text = 'exception'
                elif line.startswith('Differences '):
                    color_of_indented_text = 'normal'
                    colorize_diff = True
                else:
                    color_of_indented_text = 'normal'
                print(line)
        print()

    def print_colorized_traceback(self, formatted_traceback):
        """Report a test failure.

        ``formatted_traceback`` is a string.
        """
        for line in formatted_traceback.splitlines():
            if line.startswith('  File'):
                m = re.match(r'  File "(.*)", line (\d*), in (.*)$', line)
                if m:
                    filename, lineno, test = m.groups()
                    sys.stdout.writelines([
                        self.color('normal'), '  File "',
                        self.color('filename'), filename,
                        self.color('normal'), '", line ',
                        self.color('lineno'), lineno,
                        self.color('normal'), ', in ',
                        self.color('testname'), test,
                        self.color('normal'), '\n'])
                else:
                    print(line)
            elif line.startswith('    '):
                print(self.colorize('failed-example', line))
            elif line.startswith('Traceback (most recent call last)'):
                print(line)
            else:
                print(self.colorize('exception', line))
        print()


class FakeTest:
    """A fake test object that only has an id."""

    failureException = None

    def __init__(self, test_id):
        self._id = test_id

    def id(self):
        return self._id


# Conditional imports: we don't want zope.testrunner to have a hard
# dependency on subunit.
try:
    import subunit
    from iso8601 import UTC
    subunit.StreamResultToBytes
except (ImportError, AttributeError):
    subunit = None


# testtools is a hard dependency of subunit itself, but we guard it
# separately for richer error messages.
try:
    import testtools
    from testtools.content import Content
    from testtools.content import ContentType
    from testtools.content import content_from_file
    from testtools.content import text_content
    testtools.StreamToExtendedDecorator
except (ImportError, AttributeError):
    testtools = None


class _RunnableDecorator:
    """Permit controlling the runnable annotation on tests.

    This decorates a StreamResult, adding a setRunnable context manager to
    indicate whether a test is runnable.  (A context manager is unidiomatic
    here, but it's just about the simplest way to stuff the relevant state
    through the various layers of decorators involved without accidentally
    affecting later test results.)
    """

    def __init__(self, decorated):
        self.decorated = decorated
        self._runnable = True

    def __getattr__(self, name):
        return getattr(self.decorated, name)

    @contextmanager
    def setRunnable(self, runnable):
        orig_runnable = self._runnable
        try:
            self._runnable = runnable
            yield
        finally:
            self._runnable = orig_runnable

    def status(self, **kwargs):
        kwargs = dict(kwargs)
        kwargs['runnable'] = self._runnable
        self.decorated.status(**kwargs)


class _SortedDict(MutableMapping):
    """A dict that always returns items in sorted order.

    This differs from collections.OrderedDict in that it returns items in
    *sorted* order, not in insertion order.

    We use this as a workaround for the fact that
    testtools.ExtendedToStreamDecorator doesn't sort the details dict when
    encoding it, which makes it difficult to write stable doctests for
    subunit v2 output.
    """

    def __init__(self, items):
        self._dict = dict(items)

    def __getitem__(self, key):
        return self._dict[key]

    def __setitem__(self, key, value):
        self._dict[key] = value

    def __delitem__(self, key):
        del self._dict[key]

    def __iter__(self):
        return iter(sorted(self._dict))

    def __len__(self):
        return len(self._dict)


class SubunitOutputFormatter:
    """A subunit output formatter.

    This output formatter generates subunit-compatible output (see
    https://launchpad.net/subunit).  Subunit output is essentially a stream
    of results of unit tests.

    In this formatter, non-test events (such as layer set up) are encoded as
    specially-tagged tests.  In particular, for a layer 'foo', the fake
    tests related to layer setup and teardown are tagged with 'zope:layer'
    and are called 'foo:setUp' and 'foo:tearDown'.  Any tests within layer
    'foo' are tagged with 'zope:layer:foo'.

    Note that all tags specific to this formatter begin with 'zope:'.
    """

    # subunit output is designed for computers, so displaying a progress bar
    # isn't helpful.
    progress = False
    verbose = property(lambda self: self.options.verbose)

    TAG_INFO_SUBOPTIMAL = 'zope:info_suboptimal'
    TAG_ERROR_WITH_BANNER = 'zope:error_with_banner'
    TAG_LAYER = 'zope:layer'
    TAG_IMPORT_ERROR = 'zope:import_error'
    TAG_PROFILER_STATS = 'zope:profiler_stats'
    TAG_GARBAGE = 'zope:garbage'
    TAG_THREADS = 'zope:threads'
    TAG_REFCOUNTS = 'zope:refcounts'

    def __init__(self, options, stream=None):
        if subunit is None:
            raise Exception('Requires python-subunit 1.4.3 or better')
        if testtools is None:
            raise Exception('Requires testtools 0.9.30 or better')
        self.options = options

        if stream is None:
            stream = sys.stdout
        self._stream = stream
        self._subunit = self._subunit_factory(self._stream)

        # Used to track the last layer that was set up or torn down. Either
        # None or (layer_name, last_touched_time).
        self._last_layer = None
        self.UTC = UTC
        # Content types used in the output.
        self.TRACEBACK_CONTENT_TYPE = ContentType(
            'text', 'x-traceback', {'language': 'python', 'charset': 'utf8'})
        self.PROFILE_CONTENT_TYPE = ContentType(
            'application', 'x-binary-profile')
        self.PLAIN_TEXT = ContentType('text', 'plain', {'charset': 'utf8'})

    @classmethod
    def _subunit_factory(cls, stream):
        """Return a TestResult attached to the given stream."""
        return _RunnableDecorator(subunit.TestProtocolClient(stream))

    def _emit_timestamp(self, now=None):
        """Emit a timestamp to the subunit stream.

        If 'now' is not specified, use the current time on the system clock.
        """
        if now is None:
            now = datetime.now(self.UTC)
        self._subunit.time(now)
        return now

    def _emit_fake_test(self, message, tag, details=None):
        """Emit a successful fake test to the subunit stream.

        Use this to print tagged informative messages.
        """
        test = FakeTest(message)
        with self._subunit.setRunnable(False):
            self._subunit.startTest(test)
            self._subunit.tags([tag], [])
            self._subunit.addSuccess(test, details=details)
            self._subunit.stopTest(test)

    def _emit_error(self, error_id, tag, exc_info, runnable=False):
        """Emit an error to the subunit stream.

        Use this to pass on information about errors that occur outside of
        tests.
        """
        test = FakeTest(error_id)
        with self._subunit.setRunnable(runnable):
            self._subunit.startTest(test)
            self._subunit.tags([tag], [])
            self._subunit.addError(test, exc_info)
            self._subunit.stopTest(test)

    def _emit_failure(self, failure_id, tag, exc_info):
        """Emit an failure to the subunit stream.

        Use this to pass on information about failures that occur outside of
        tests.
        """
        test = FakeTest(failure_id)
        self._subunit.addFailure(test, exc_info)

    def _enter_layer(self, layer_name):
        """Tell subunit that we are entering a layer."""
        self._subunit.tags([f'zope:layer:{layer_name}'], [])

    def _exit_layer(self, layer_name):
        """Tell subunit that we are exiting a layer."""
        self._subunit.tags([], [f'zope:layer:{layer_name}'])

    def info(self, message):
        """Print an informative message."""
        # info() output is not relevant to actual test results.  It only
        # says things like "Running tests" or "Tearing down left over
        # layers", things that are communicated already by the subunit
        # stream.  Just suppress the info() output.
        pass

    def info_suboptimal(self, message):
        """Print an informative message about losing some of the features.

        For example, when you run some tests in a subprocess, you lose the
        ability to use the debugger.
        """
        # Used _only_ to indicate running in a subprocess.
        self._emit_fake_test(message.strip(), self.TAG_INFO_SUBOPTIMAL)

    def error(self, message):
        """Report an error."""
        # XXX: Mostly used for user errors, sometimes used for errors in the
        # test framework, sometimes used to record layer setUp failure (!!!).
        self._stream.write(f'{message}\n')

    def error_with_banner(self, message):
        """Report an error with a big ASCII banner."""
        # Either "Could not communicate with subprocess"
        # Or "Can't post-mortem debug when running a layer as a subprocess!"
        self._emit_fake_test(message, self.TAG_ERROR_WITH_BANNER)

    def profiler_stats(self, stats):
        """Report profiler stats."""
        fd, filename = tempfile.mkstemp(prefix='zope.testrunner-')
        os.close(fd)
        try:
            stats.dump_stats(filename)
            profile_content = content_from_file(
                filename, content_type=self.PROFILE_CONTENT_TYPE)
            details = {'profiler-stats': profile_content}
            # Name the test 'zope:profiler_stats' just like its tag.
            self._emit_fake_test(
                self.TAG_PROFILER_STATS, self.TAG_PROFILER_STATS, details)
        finally:
            os.unlink(filename)

    def import_errors(self, import_errors):
        """Report test-module import errors (if any)."""
        if import_errors:
            for error in import_errors:
                self._emit_error(
                    error.module, self.TAG_IMPORT_ERROR, error.exc_info,
                    runnable=True)

    def tests_with_errors(self, errors):
        """Report names of tests with errors (if any).

        Simply not supported by the subunit formatter. Fancy summary output
        doesn't make sense.
        """
        pass

    def tests_with_failures(self, failures):
        """Report names of tests with failures (if any).

        Simply not supported by the subunit formatter. Fancy summary output
        doesn't make sense.
        """
        pass

    def modules_with_import_problems(self, import_errors):
        """Report names of modules with import problems (if any)."""
        # This is simply a summary method, and subunit output doesn't
        # benefit from summaries.
        pass

    def summary(self, n_tests, n_failures, n_errors, n_seconds,
                n_skipped=0):
        """Summarize the results of a single test layer.

        Since subunit is a stream protocol format, it has no need for a
        summary. When the stream is finished other tools can generate a
        summary if so desired.
        """
        pass

    def totals(self, n_tests, n_failures, n_errors, n_seconds, n_skipped=0):
        """Summarize the results of all layers.

        Simply not supported by the subunit formatter. Fancy summary output
        doesn't make sense.
        """
        pass

    def _emit_exists(self, test):
        """Emit an indication that a test exists.

        With the v1 protocol, we just emit a fake success line.
        """
        self._subunit.addSuccess(test)

    def list_of_tests(self, tests, layer_name):
        """Report a list of test names."""
        self._enter_layer(layer_name)
        for test in tests:
            self._subunit.startTest(test)
            self._emit_exists(test)
            self._subunit.stopTest(test)
        self._exit_layer(layer_name)

    def garbage(self, garbage):
        """Report garbage generated by tests."""
        # XXX: Really, 'garbage', 'profiler_stats' and the 'refcounts' twins
        # ought to add extra details to a fake test that represents the
        # summary information for the whole suite. However, there's no event
        # on output formatters for "everything is really finished, honest". --
        # jml, 2010-02-14
        details = {'garbage': text_content(str(garbage))}
        self._emit_fake_test(self.TAG_GARBAGE, self.TAG_GARBAGE, details)

    def test_garbage(self, test, garbage):
        """Report garbage generated by a test.

        Encoded in the subunit stream as a test error.  Clients can filter
        out these tests based on the tag if they don't think garbage should
        fail the test run.
        """
        # XXX: Perhaps 'test_garbage' and 'test_threads' ought to be within
        # the output for the actual test, appended as details to whatever
        # result the test gets. Not an option with the present API, as there's
        # no event for "no more output for this test". -- jml, 2010-02-14
        self._subunit.startTest(test)
        self._subunit.tags([self.TAG_GARBAGE], [])
        self._subunit.addError(
            test, details={'garbage': text_content(str(garbage))})
        self._subunit.stopTest(test)

    def test_threads(self, test, new_threads):
        """Report threads left behind by a test.

        Encoded in the subunit stream as a test error.  Clients can filter
        out these tests based on the tag if they don't think left-over
        threads should fail the test run.
        """
        self._subunit.startTest(test)
        self._subunit.tags([self.TAG_THREADS], [])
        self._subunit.addError(
            test, details={'threads': text_content(str(new_threads))})
        self._subunit.stopTest(test)

    def test_cycles(self, test, cycles):
        """Report cycles left behind by a test."""
        pass  # not implemented

    def refcounts(self, rc, prev):
        """Report a change in reference counts."""
        details = _SortedDict({
            'sys-refcounts': text_content(str(rc)),
            'changes': text_content(str(rc - prev)),
        })
        # XXX: Emit the details dict as JSON?
        self._emit_fake_test(self.TAG_REFCOUNTS, self.TAG_REFCOUNTS, details)

    def detailed_refcounts(self, track, rc, prev):
        """Report a change in reference counts, with extra detail."""
        details = _SortedDict({
            'sys-refcounts': text_content(str(rc)),
            'changes': text_content(str(rc - prev)),
            'track': text_content(str(track.delta)),
        })
        self._emit_fake_test(self.TAG_REFCOUNTS, self.TAG_REFCOUNTS, details)

    def start_set_up(self, layer_name):
        """Report that we're setting up a layer.

        We do this by emitting a fake test of the form '$LAYER_NAME:setUp'
        and adding a tag of the form 'zope:layer:$LAYER_NAME' to the current
        tag context.

        The next output operation should be stop_set_up().
        """
        test = FakeTest(f'{layer_name}:setUp')
        now = self._emit_timestamp()
        with self._subunit.setRunnable(False):
            self._subunit.startTest(test)
            self._subunit.tags([self.TAG_LAYER], [])
        self._last_layer = (layer_name, now)

    def stop_set_up(self, seconds):
        """Report that we've set up a layer.

        Should be called right after start_set_up().
        """
        layer_name, start_time = self._last_layer
        self._last_layer = None
        test = FakeTest(f'{layer_name}:setUp')
        self._emit_timestamp(start_time + timedelta(seconds=seconds))
        with self._subunit.setRunnable(False):
            self._subunit.addSuccess(test)
            self._subunit.stopTest(test)
        self._enter_layer(layer_name)

    def layer_failure(self, failure_type, exc_info):
        layer_name, start_time = self._last_layer
        self._emit_failure(
            f'{layer_name}:{failure_type}', self.TAG_LAYER, exc_info)

    def start_tear_down(self, layer_name):
        """Report that we're tearing down a layer.

        We do this by emitting a fake test of the form
        '$LAYER_NAME:tearDown' and removing a tag of the form
        'layer:$LAYER_NAME' from the current tag context.

        The next output operation should be stop_tear_down() or
        tear_down_not_supported().
        """
        test = FakeTest(f'{layer_name}:tearDown')
        self._exit_layer(layer_name)
        now = self._emit_timestamp()
        with self._subunit.setRunnable(False):
            self._subunit.startTest(test)
            self._subunit.tags([self.TAG_LAYER], [])
        self._last_layer = (layer_name, now)

    def stop_tear_down(self, seconds):
        """Report that we've torn down a layer.

        Should be called right after start_tear_down().
        """
        layer_name, start_time = self._last_layer
        self._last_layer = None
        test = FakeTest(f'{layer_name}:tearDown')
        self._emit_timestamp(start_time + timedelta(seconds=seconds))
        with self._subunit.setRunnable(False):
            self._subunit.addSuccess(test)
            self._subunit.stopTest(test)

    def tear_down_not_supported(self):
        """Report that we could not tear down a layer.

        Should be called right after start_tear_down().
        """
        layer_name, start_time = self._last_layer
        self._last_layer = None
        test = FakeTest(f'{layer_name}:tearDown')
        self._emit_timestamp()
        with self._subunit.setRunnable(False):
            self._subunit.addSkip(test, 'tearDown not supported')
            self._subunit.stopTest(test)

    def start_test(self, test, tests_run, total_tests):
        """Report that we're about to run a test.

        The next output operation should be test_success(), test_error(), or
        test_failure().
        """
        self._emit_timestamp()
        self._subunit.startTest(test)

    def test_success(self, test, seconds):
        """Report that a test was successful.

        Should be called right after start_test().

        The next output operation should be stop_test().
        """
        self._emit_timestamp()
        self._subunit.addSuccess(test)

    def test_skipped(self, test, reason):
        """Report that a test was skipped.

        Should be called right after start_test().

        The next output operation should be stop_test().
        """
        self._subunit.addSkip(test, reason)

    def _exc_info_to_details(self, exc_info):
        """Translate 'exc_info' into a details dict usable with subunit."""
        # In an ideal world, we'd use the pre-bundled 'TracebackContent'
        # class from testtools.  However, 'OutputFormatter' contains special
        # logic to handle errors from doctests, so we have to use that and
        # manually create an object equivalent to an instance of
        # 'TracebackContent'.
        formatter = OutputFormatter(None)
        traceback = formatter.format_traceback(exc_info)

        # We have no idea if the traceback is a str object or a
        # bytestring with non-ASCII characters.  We had best be careful when
        # handling it.
        if isinstance(traceback, bytes):
            # Assume the traceback was UTF-8-encoded, but still be careful.
            str_tb = traceback.decode('utf-8', 'replace')
        else:
            str_tb = traceback

        return _SortedDict({
            'traceback': Content(
                self.TRACEBACK_CONTENT_TYPE,
                lambda: [str_tb.encode('utf8')]),
        })

    def _add_std_streams_to_details(self, details, stdout, stderr):
        """Add buffered standard stream contents to a subunit details dict."""
        if stdout:
            if isinstance(stdout, bytes):
                stdout = stdout.decode('utf-8', 'replace')
            details['test-stdout'] = Content(
                self.PLAIN_TEXT, lambda: [stdout.encode('utf-8')])
        if stderr:
            if isinstance(stderr, bytes):
                stderr = stderr.decode('utf-8', 'replace')
            details['test-stderr'] = Content(
                self.PLAIN_TEXT, lambda: [stderr.encode('utf-8')])

    def test_error(self, test, seconds, exc_info, stdout=None, stderr=None):
        """Report that an error occurred while running a test.

        Should be called right after start_test().

        The next output operation should be stop_test().
        """
        self._emit_timestamp()
        details = self._exc_info_to_details(exc_info)
        self._add_std_streams_to_details(details, stdout, stderr)
        self._subunit.addError(test, details=details)

    def test_failure(self, test, seconds, exc_info, stdout=None, stderr=None):
        """Report that a test failed.

        Should be called right after start_test().

        The next output operation should be stop_test().
        """
        self._emit_timestamp()
        details = self._exc_info_to_details(exc_info)
        self._add_std_streams_to_details(details, stdout, stderr)
        self._subunit.addFailure(test, details=details)

    def stop_test(self, test, gccount):
        """Clean up the output state after a test."""
        self._subunit.stopTest(test)

    def stop_tests(self):
        """Clean up the output state after a collection of tests."""
        # subunit handles all of this itself.
        pass


class SubunitV2OutputFormatter(SubunitOutputFormatter):
    """A subunit v2 output formatter."""

    @classmethod
    def _subunit_factory(cls, stream):
        """Return a TestResult attached to the given stream."""
        stream_result = _RunnableDecorator(subunit.StreamResultToBytes(stream))
        result = testtools.ExtendedToStreamDecorator(stream_result)
        # Lift our decorating method up so that we can get at it easily.
        result.setRunnable = stream_result.setRunnable
        result.startTestRun()
        return result

    def error(self, message):
        """Report an error."""
        # XXX: Mostly used for user errors, sometimes used for errors in the
        # test framework, sometimes used to record layer setUp failure (!!!).
        self._subunit.status(
            file_name='error', file_bytes=str(message).encode('utf-8'),
            eof=True, mime_type=repr(self.PLAIN_TEXT))

    def _emit_exists(self, test):
        """Emit an indication that a test exists."""
        now = datetime.now(self.UTC)
        self._subunit.status(
            test_id=test.id(), test_status='exists',
            test_tags=self._subunit.current_tags, timestamp=now)


@dataclass
class TestSuiteInfo:

    testCases: list = field(default_factory=list)
    errors: int = 0
    failures: int = 0
    time: float = 0.0

    @property
    def tests(self):
        return len(self.testCases)

    @property
    def successes(self):
        return self.tests - self.errors - self.failures


@dataclass
class TestCaseInfo:

    test: object
    time: float
    testClassName: str
    testName: str
    failure: bool = None
    error: bool = None


def get_test_class_name(test):
    """Compute the test class name from the test object."""
    return f'{test.__module__}.{test.__class__.__name__}'


def filename_to_suite_name_parts(filename):
    # lop off whatever portion of the path we have in common
    # with the current working directory; crude, but about as
    # much as we can do :(
    filenameParts = Path(filename).parts
    cwdParts = Path.cwd().parts
    longest = min(len(filenameParts), len(cwdParts))
    for i in range(longest):
        if filenameParts[i] != cwdParts[i]:
            break

    if i < len(filenameParts) - 1:

        # The real package name couldn't have a '.' in it. This
        # makes sense for the common egg naming patterns, and
        # will still work in other cases

        suiteNameParts = []
        for part in reversed(filenameParts[i:-1]):
            if '.' in part:
                break
            suiteNameParts.insert(0, part)

        # don't lose the filename, which would have a . in it
        suiteNameParts.append(filenameParts[-1])
        return suiteNameParts


def parse_doc_file_case(test):
    if not isinstance(test, doctest.DocFileCase):
        return None, None, None

    filename = test._dt_test.filename
    suiteNameParts = filename_to_suite_name_parts(filename)
    testSuite = f'doctest-{"-".join(suiteNameParts)}'
    testName = test._dt_test.name
    testClassName = '.'.join(suiteNameParts[:-1])
    return testSuite, testName, testClassName


def parse_doc_test_case(test):
    if not isinstance(test, doctest.DocTestCase):
        return None, None, None

    testDottedNameParts = test._dt_test.name.split('.')
    testClassName = get_test_class_name(test)
    testSuite = testClassName = '.'.join(testDottedNameParts[:-1])
    testName = testDottedNameParts[-1]
    return testSuite, testName, testClassName


def parse_manuel(test):
    if not (HAVE_MANUEL and isinstance(test, manuel.testing.TestCase)):
        return None, None, None
    filename = test.regions.location
    suiteNameParts = filename_to_suite_name_parts(filename)
    testSuite = f'manuel-{"-".join(suiteNameParts)}'
    testName = suiteNameParts[-1]
    testClassName = '.'.join(suiteNameParts[:-1])
    return testSuite, testName, testClassName


def parse_startup_failure(test):
    if not isinstance(test, StartUpFailure):
        return None, None, None
    testModuleName = test.module
    return testModuleName, 'Startup', testModuleName


def parse_unittest(test):
    testId = test.id()
    if testId is None:
        return None, None, None
    testClassName = get_test_class_name(test)
    testSuite = testClassName
    testName = testId[len(testClassName) + 1:]
    return testSuite, testName, testClassName


class XMLOutputFormattingWrapper:
    """Output formatter which delegates to another formatter for all
    operations, but also prepares an element tree of test output.
    """

    def __init__(self, delegate, folder):
        self.delegate = delegate
        self._testSuites = {}  # test class -> list of test names
        self.folder = folder

    def __getattr__(self, name):
        return getattr(self.delegate, name)

    def test_failure(self, test, seconds, exc_info, stdout=None, stderr=None):
        self._record(test, seconds, failure=exc_info)
        # stdout and stderr are only passed into us when using buffering
        # (--buffer). self.delegate also comes from zope.testrunner then.
        if stdout is None and stderr is None:
            # the normal case
            return self.delegate.test_failure(test, seconds, exc_info)
        return self.delegate.test_failure(
            test, seconds, exc_info, stdout=stdout, stderr=stderr
        )

    def test_error(self, test, seconds, exc_info, stdout=None, stderr=None):
        self._record(test, seconds, error=exc_info)
        if stdout is None and stderr is None:
            # the normal case
            return self.delegate.test_error(test, seconds, exc_info)
        return self.delegate.test_error(
            test, seconds, exc_info, stdout=stdout, stderr=stderr
        )

    def test_success(self, test, seconds):
        self._record(test, seconds)
        return self.delegate.test_success(test, seconds)

    def import_errors(self, import_errors):
        if import_errors:
            for test in import_errors:
                self._record(test, 0, error=test.exc_info)
        return self.delegate.import_errors(import_errors)

    def _record(self, test, seconds, failure=None, error=None):
        for parser in [parse_doc_file_case,
                       parse_doc_test_case,
                       parse_manuel,
                       parse_startup_failure,
                       parse_unittest]:
            testSuite, testName, testClassName = parser(test)
            if (testSuite, testName, testClassName) != (None, None, None):
                break

        if (testSuite, testName, testClassName) == (None, None, None):
            raise TypeError(
                'Unknown test type: Could not compute testSuite, testName,'
                f' testClassName: {test!r}'
            )

        suite = self._testSuites.setdefault(testSuite, TestSuiteInfo())
        suite.testCases.append(TestCaseInfo(
            test, seconds, testClassName, testName, failure, error))

        if failure is not None:
            suite.failures += 1

        if error is not None:
            suite.errors += 1

        if seconds:
            suite.time += seconds

    def writeXMLReports(self, properties={}):

        timestamp = datetime.now().isoformat()
        hostname = socket.gethostname()

        reportsDir = self.folder / 'testreports'
        reportsDir.mkdir(exist_ok=True)

        for name, suite in self._testSuites.items():
            filename = reportsDir / f'{name}.xml'

            testSuiteNode = ElementTree.Element('testsuite')

            testSuiteNode.set('tests', str(suite.tests))
            testSuiteNode.set('errors', str(suite.errors))
            testSuiteNode.set('failures', str(suite.failures))
            testSuiteNode.set('hostname', hostname)
            testSuiteNode.set('name', name)
            testSuiteNode.set('time', str(suite.time))
            testSuiteNode.set('timestamp', timestamp)

            propertiesNode = ElementTree.Element('properties')
            testSuiteNode.append(propertiesNode)

            for k, v in properties.items():
                propertyNode = ElementTree.Element('property')
                propertiesNode.append(propertyNode)

                propertyNode.set('name', k)
                propertyNode.set('value', v)

            for testCase in suite.testCases:
                testCaseNode = ElementTree.Element('testcase')
                testSuiteNode.append(testCaseNode)

                testCaseNode.set('classname', testCase.testClassName)
                testCaseNode.set('name', testCase.testName)
                testCaseNode.set('time', str(testCase.time))

                if testCase.error:
                    errorNode = ElementTree.Element('error')
                    testCaseNode.append(errorNode)

                    try:
                        excType, excInstance, tb = testCase.error
                        errorMessage = str(excInstance)
                        stackTrace = ''.join(traceback.format_tb(tb))
                    finally:  # Avoids a memory leak
                        del tb

                    errorNode.set('message', errorMessage.split('\n')[0])
                    errorNode.set('type', str(excType))
                    text = (errorMessage + '\n\n' + stackTrace)
                    errorNode.text = text

                if testCase.failure:

                    failureNode = ElementTree.Element('failure')
                    testCaseNode.append(failureNode)

                    try:
                        excType, excInstance, tb = testCase.failure
                        errorMessage = str(excInstance)
                        stackTrace = ''.join(traceback.format_tb(tb))
                    except UnicodeEncodeError:
                        errorMessage = 'Could not extract error str ' \
                            'for unicode error'
                        stackTrace = ''.join(traceback.format_tb(tb))
                    finally:  # Avoids a memory leak
                        del tb

                    failureNode.set('message', errorMessage.split('\n')[0])
                    failureNode.set('type', str(excType))
                    text = f'{errorMessage}\n\n{stackTrace}'
                    failureNode.text = text

            # We don't have a good way to capture these yet, so they are empty:
            systemOutNode = ElementTree.Element('system-out')
            testSuiteNode.append(systemOutNode)
            systemErrNode = ElementTree.Element('system-err')
            testSuiteNode.append(systemErrNode)

            # indent the XML structure
            with suppress(AttributeError):
                ElementTree.indent(testSuiteNode)
            text = ElementTree.tostring(testSuiteNode).decode('utf-8')

            # Write file
            with open(filename, 'w') as outputFile:
                outputFile.write(text)



##############################################################################
#
# Copyright (c) 2004-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Generic features for the test runner.
"""

import zope.interface

import zope.testrunner.interfaces


@zope.interface.implementer(zope.testrunner.interfaces.IFeature)
class Feature:
    """A base class implementing no-op methods for the IFeature interface."""

    active = False

    def __init__(self, runner):
        self.runner = runner

    def global_setup(self):
        """Executed once when the test runner is being set up."""
        pass

    def late_setup(self):
        """Executed once right before the actual tests get executed and after
        all global setups have happened.
        """
        pass

    def layer_setup(self, layer):
        """Executed once after a layer was set up."""
        pass

    def layer_teardown(self, layer):
        """Executed once after a layer was run."""
        pass

    def test_setup(self):
        """Executed once before each test."""
        pass

    def test_teardown(self):
        """Executed once after each test."""
        pass

    def early_teardown(self):
        """Executed once directly after all tests."""
        pass

    def global_teardown(self):
        """Executed once after all tests where run and early teardowns have
        happened."""
        pass

    def report(self):
        """Executed once after all tests have been run and all setup was
        torn down."""
        pass



##############################################################################
#
# Copyright (c) 2017 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unit tests for test discovery."""

import doctest
import os.path
import unittest

from zope.testrunner import find


class UniquenessOptions:
    """A basic mock of our command-line options."""

    keepbytecode = True
    at_level = 99
    only_level = None
    test = []
    module = []
    require_unique_ids = True


class TestUniqueness(unittest.TestCase):
    """Test how the testrunner handles non-unique IDs."""

    def setUp(self):
        super().setUp()
        suites = [
            doctest.DocFileSuite('testrunner-ex/sampletests.rst'),
            doctest.DocFileSuite('testrunner-ex/sampletests.rst'),
            doctest.DocFileSuite('testrunner-ex/sampletestsl.rst'),
            doctest.DocFileSuite('testrunner-ex/sampletestsl.rst'),
        ]
        self.test_suites = unittest.TestSuite(suites)

    def test_tests_from_suite_records_duplicate_test_ids(self):
        # If tests_from_suite encounters a test ID which has already been
        # registered, it records them in the duplicated_test_ids set it is
        # passed.
        duplicated_test_ids = set()
        list(find.tests_from_suite(
            self.test_suites, UniquenessOptions(),
            duplicated_test_ids=duplicated_test_ids))
        self.assertNotEqual(0, len(duplicated_test_ids))

    def test_tests_from_suite_ignores_duplicate_ids_if_option_not_set(self):
        # If the require_unique_ids option is not set, tests_from_suite will
        # not record any of the IDs as duplicates.
        options = UniquenessOptions()
        options.require_unique_ids = False
        duplicated_test_ids = set()
        list(find.tests_from_suite(
            self.test_suites, options,
            duplicated_test_ids=duplicated_test_ids))
        self.assertEqual(0, len(duplicated_test_ids))

    def test_find_tests_raises_error_if_duplicates_found(self):
        # If find_tests, which calls tests_from_suite, finds a duplicate
        # test ID, it raises an error.
        self.assertRaises(
            find.DuplicateTestIDError,
            find.find_tests, UniquenessOptions(), [self.test_suites])

    def test_DuplicateTestIDError_message_contains_all_test_ids(self):
        # The error raised by find_tests when duplicates are encountered
        # contains all the duplicate test IDs it found.
        with self.assertRaises(find.DuplicateTestIDError) as e:
            find.find_tests(UniquenessOptions(), [self.test_suites])
        self.assertIn(
            os.path.join('testrunner-ex', 'sampletests.rst'), str(e.exception))
        self.assertIn(
            os.path.join('testrunner-ex', 'sampletestsl.rst'),
            str(e.exception))


class TestIdentifierMatches(unittest.TestCase):
    """Test which folders are ignored by the test runner."""

    def test_ignored(self):
        folders = ('spaces are bad', 'dashes-bad', 'ünicodeLeading')
        for name in folders:
            self.assertFalse(find.identifier(name), f'{name} is accepted')

    def test_accepted(self):
        folders = (
            'random', 'CamelCase', 'Under_scores', '_leading_scores', 'unicöde'
        )
        for name in folders:
            self.assertTrue(find.identifier(name), f'{name} is not accepted')



import os
import unittest


class Layer1:
    pass


class Layer2:
    pass


class Test1(unittest.TestCase):
    layer = Layer1

    def test_that_chdirs(self):
        os.chdir(os.path.dirname(__file__))


class Test2(unittest.TestCase):
    layer = Layer2

    def test(self):
        pass



##############################################################################
#
# Copyright (c) 2004-2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Shuffle tests discovered before executing them.
"""

import math
import random
import time

import zope.testrunner.feature


class Shuffle(zope.testrunner.feature.Feature):
    """Take the tests found so far and shuffle them."""

    def __init__(self, runner):
        super().__init__(runner)
        self.active = runner.options.shuffle
        self.seed = runner.options.shuffle_seed
        if self.seed is None:
            # We can't rely on the random modules seed initialization because
            # we can't introspect the seed later for reporting.  This is a
            # simple emulation of what random.Random.seed does anyway.
            self.seed = int(time.time() * 256)  # use fractional seconds

    def global_setup(self):
        rng = random.Random(self.seed)
        # in case somebody tries to use a string as the seed
        rng.seed(self.seed, version=1)
        # Be careful to shuffle the layers in a deterministic order!
        for layer, suite in sorted(self.runner.tests_by_layer_name.items()):
            # Test suites cannot be modified through a public API.  We thus
            # take a mutable copy of the list of tests of that suite, shuffle
            # that and replace the test suite instance with a new one of the
            # same class.
            tests = list(suite)
            # The standard library guarantees rng.random() will return the
            # same floats if given the same seed.  It makes no guarantees
            # for rng.randrange, or rng.choice, or rng.shuffle.  Experiments
            # show that Python happily breaks backwards compatibility for
            # these functions.  We used to use `rng.shuffle(tests,
            # random=rng.random)` to trick it into using the old algorithm,
            # but even this was deprecated in Python 3.9 and removed in
            # 3.11.  Accordingly, inline equivalent code which we know is
            # stable across Python versions.
            floor = math.floor
            for i in reversed(range(1, len(tests))):
                # Pick an element in tests[:i+1] with which to exchange
                # tests[i].
                j = floor(rng.random() * (i + 1))
                tests[i], tests[j] = tests[j], tests[i]
            self.runner.tests_by_layer_name[layer] = suite.__class__(tests)

    def report(self):
        msg = "Tests were shuffled using seed number %d." % self.seed
        self.runner.options.output.info(msg)



##############################################################################
#
# Copyright (c) 2023 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Unit tests for the XML reports
"""
import contextlib
import io
import shutil
import sys
import tempfile
import unittest
from pathlib import Path

from zope import testrunner


class Base(unittest.TestCase):

    def tearDown(self):
        self._cleanup(self.tmpdir)
        return super().tearDown()

    def _cleanup(self, path):
        assert path.exists()
        shutil.rmtree(path)


class TestXMLOutput(Base):
    """Test the XML output of `zope.testrunner`

    By default, `zope.testrunner` reports the results to the terminal.
    CI systems however, need a machine readable format
    to actually parse the results and display them.

    The `JUnit` package from the Java community, long ago,
    created an XML report format that since then
    every CI system is able to read.

    The `--xml` option tells the test runner to produce such a report.
    """

    def setUp(self):
        self.tmpdir = Path(tempfile.mkdtemp())
        directory_with_tests = Path(Path(__file__).parent, 'testrunner-ex')
        self.arg_defaults = [
            '--path', str(directory_with_tests),
            '--tests-pattern', '^sampletestsf?$',
        ]
        self.default_argv = f'test -u --xml={self.tmpdir} -t sample3'.split()
        self.reports_folder = self.tmpdir / 'testreports'

    def _run_tests(self):
        stream1 = io.StringIO()
        stream2 = io.StringIO()
        with contextlib.redirect_stdout(stream1):
            with contextlib.redirect_stderr(stream2):
                testrunner.run_internal(self.arg_defaults, self.default_argv)

    def test_xml_report(self):
        self._run_tests()
        # A report has been created:
        self.assertTrue(self.reports_folder.exists())

        # There are 5 reports:
        reports = [x for x in self.reports_folder.iterdir()]
        self.assertEqual(len(reports), 5)

    def test_xml_report_details(self):
        """Let's look at an individual report."""
        self._run_tests()
        report = self.reports_folder / 'sample3.sampletests.TestA.xml'

        # 3 tests were run with no errors nor failures:
        content = report.read_text()
        self.assertIn(' tests="3" ', content)
        self.assertIn(' errors="0" ', content)
        self.assertIn(' failures="0" ', content)

        # The specific tests that were run:
        self.assertIn(
            '<testcase classname="sample3.sampletests.TestA" name="test_x1"',
            content)
        self.assertIn(
            '<testcase classname="sample3.sampletests.TestA" name="test_y0"',
            content)
        self.assertIn(
            '<testcase classname="sample3.sampletests.TestA" name="test_z0"',
            content)


class TextXMLOutputWithErrors(Base):
    """If errors/failures happen, they are also reported in the XML files

    Code copied and adapted from `testrunner-errors.rst`
    """

    def setUp(self):
        self.tmpdir = Path(tempfile.mkdtemp())
        directory_with_tests = self.tmpdir / 'testrunner-ex'
        self.reports_folder = self.tmpdir / 'testreports'

        source = Path(__file__).parent / 'testrunner-ex'
        shutil.copytree(source, directory_with_tests)

        self.defaults = [
            '--path', str(directory_with_tests),
            '--tests-pattern', '^sampletestsf?$',
            f'--xml={self.tmpdir}',
        ]

    def _run_tests(self):
        stream1 = io.StringIO()
        stream2 = io.StringIO()
        with contextlib.redirect_stdout(stream1):
            with contextlib.redirect_stderr(stream2):
                testrunner.run_internal(self.defaults)

    def test_xml_report_with_errors(self):
        sys.argv = 'test --tests-pattern ^sampletests(f|_e|_f)?$ '.split()
        self._run_tests()

        self.assertTrue(self.reports_folder.exists())
        self.assertEqual(len([x for x in self.reports_folder.iterdir()]), 106)

    def test_xml_report_with_errors_details(self):
        sys.argv = 'test --tests-pattern ^sampletests(f|_e|_f)?$ '.split()
        self._run_tests()
        report = self.reports_folder / 'sample2.sampletests_e.Test.xml'
        self.assertTrue(report.exists())

        content = report.read_text()
        self.assertIn(' tests="5" ', content)
        self.assertIn(' errors="1" ', content)
        self.assertIn(' failures="0" ', content)

        # The error message is reported:
        self.assertIn('<error message="name \'y\' is not defined"', content)

        # And the traceback is there as well:
        self.assertIn('sampletests_e.py", line 47, in test3', content)

    def test_xml_report_with_failures(self):
        sys.argv = 'test --tests-pattern ^sampletests(f|_e|_f)?$ '.split()
        self._run_tests()
        report = (
            self.reports_folder /
            'sample2.sampletests_f.Test.xml'
        )
        self.assertTrue(report.exists())

        content = report.read_text()
        self.assertIn(' tests="1" ', content)
        self.assertIn(' errors="0" ', content)
        self.assertIn(' failures="1" ', content)

        # The failure is reported:
        self.assertIn("class 'AssertionError'", content)
        self.assertIn("self.assertEqual(1, 0)", content)
